#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "breakpoint.h"
#include "io.h"
#include "memory.h"
#include "osd.h"
#include "schedule.h"
#include "z80.h"


#ifdef USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
//------------------------------------------------------
//  ץɥ饹
//------------------------------------------------------
////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
cWndMem::cWndMem( VM6 *vm, const P6ID& id ) : P6DEVICE(vm,id){}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
cWndMem::~cWndMem( void ){}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
SDL_Surface *cWndMem::Init( void )
{
	// ɽɥ쥹
	Addr = 0;
	
	return ZCons::Create( MEMWINW, MEMWINH, "MEMORY" );
}


////////////////////////////////////////////////////////////////
// ɥ
////////////////////////////////////////////////////////////////
void cWndMem::Update( void )
{
	WORD addr = Addr;
	static Uint8 TexCol[] = { FC_GRAY, FC_BLUE, FC_GREEN, FC_CYAN, FC_RED, FC_MAGENTA, FC_YELLOW, FC_WHITE };
	int i,j;
	
	ZCons::Locate( 0, 0 );
	
	// Read ֥å̾ɽ
	ZCons::Printf( "R" );
	for( i=0; i<4; i++ ){
		ZCons::SetColor( TexCol[i] );
		ZCons::Printf( " %-8s", vm->mem->GetReadMemBlk( i ) );
	}
	ZCons::Printf( "\n " );
	for( i=4; i<8; i++ ){
		ZCons::SetColor( TexCol[i] );
		ZCons::Printf( " %-8s", vm->mem->GetReadMemBlk( i ) );
	}
	ZCons::Printf( "\nW" );
	
	// Write ֥å̾ɽ
	for( i=0; i<4; i++ ){
		ZCons::SetColor( TexCol[i] );
		ZCons::Printf( " %-8s", vm->mem->GetWriteMemBlk( i ) );
	}
	ZCons::Printf( "\n " );
	for( i=4; i<8; i++ ){
		ZCons::SetColor( TexCol[i] );
		ZCons::Printf( " %-8s", vm->mem->GetWriteMemBlk( i ) );
	}
	ZCons::Printf( "\n" );
	
	// ɽ
	for( i=0; i<ZCons::GetYline()-5; i++ ){
		ZCons::Printf( "\n" );
		ZCons::SetColor( TexCol[ addr>>13 ] );
		ZCons::Printf( "%04X:", addr );
		ZCons::SetColor( FC_WHITE );
		for( j=0; j<8; j++)
			ZCons::Printf( "%02X ", vm->mem->Read(addr+j) );
		for( j=0; j<8; j++)
			ZCons::PutCharH( vm->mem->Read(addr+j) );
		addr += 8;
	}
}


//------------------------------------------------------
//  쥸ɥ饹
//------------------------------------------------------
////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
cWndReg::cWndReg( VM6 *vm, const P6ID& id ) : P6DEVICE(vm,id){}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
cWndReg::~cWndReg( void ){}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
SDL_Surface *cWndReg::Init( void )
{
	return ZCons::Create( REGWINW, REGWINH, "REGISTER" );
}


////////////////////////////////////////////////////////////////
// ɥ
////////////////////////////////////////////////////////////////
void cWndReg::Update( void )
{
	const char flags[9] = "SZ.H.PNC";
	char fbuf[9];
	Register reg;
	char DisCode[128];
	int i,j;
	
	// 쥸ͼ
	vm->cpu->GetRegister( &reg );
	
	// ե饰ξּ
	for( j=0,i=reg.AF.B.l; j<8; j++,i<<=1 ) fbuf[j] = i&0x80 ? flags[j] : '.';
	fbuf[8]='\0';
	
	// 1饤ե֥
	vm->cpu->Disasm( DisCode, reg.PC.W );
	
	ZCons::Locate( 0, 0 ); ZCons::Print( "AF :%04X  BC :%04X  DE :%04X  HL :%04X", reg.AF.W, reg.BC.W, reg.DE.W, reg.HL.W );
	ZCons::Locate( 0, 1 ); ZCons::Print( "AF':%04X  BC':%04X  DE':%04X  HL':%04X", reg.AF1.W, reg.BC1.W, reg.DE1.W, reg.HL1.W );
	ZCons::Locate( 0, 2 ); ZCons::Print( "IX :%04X  IY :%04X  PC :%04X  SP :%04X", reg.IX.W, reg.IY.W, reg.PC.W, reg.SP.W );
	ZCons::Locate( 0, 3 ); ZCons::Print( "FLAG:[%s] I:%02X IFF:%d IM:%1d HALT:%1d",  fbuf, reg.I, reg.IFF, reg.IM, reg.Halt );
	ZCons::Locate( 0, 4 ); ZCons::Print( "--------------------------------------" );
	ZCons::Locate( 0, 5 ); ZCons::Print( " %-36s", DisCode );
}


//------------------------------------------------------
//  ˥ɥ饹
//------------------------------------------------------

////////////////////////////////////////////////////////////////
// ̿μȽơ֥
////////////////////////////////////////////////////////////////
enum MonitorJob	// 
{
	MONITOR_NONE = 0,
	
	MONITOR_HELP,
	
	MONITOR_GO,
	MONITOR_TRACE,
	MONITOR_STEP,
	MONITOR_STEPALL,
	MONITOR_BREAK,
	
	MONITOR_READ,
	MONITOR_WRITE,
	MONITOR_FILL,
	MONITOR_MOVE,
	MONITOR_SEARCH,
	MONITOR_OUT,
	MONITOR_LOADMEM,
	MONITOR_SAVEMEM,
	
	MONITOR_RESET,
	MONITOR_REG,
	MONITOR_DISASM,
	
	EndofMONITOR
};

struct{
	int Step;
	char *cmd;
	char *HelpMes;
}MonitorCmd[]=
	{
	{ MONITOR_HELP,		"help",		"إפɽ" },
	{ MONITOR_HELP,		"?",		"    " },
	{ MONITOR_GO,		"go",		"¹" },
	{ MONITOR_GO,		"g",		"    " },
	{ MONITOR_TRACE,	"trace",	"ȥ졼¹" },
	{ MONITOR_TRACE,	"t",		"    " },
	{ MONITOR_STEP,		"step",		"ƥå׼¹" },
	{ MONITOR_STEP,		"s",		"    " },
	{ MONITOR_STEPALL,	"S",		"    " },
	{ MONITOR_BREAK,	"break",	"֥졼ݥ" },
	{ MONITOR_BREAK,	"b",		"    " },
	{ MONITOR_READ,		"read",		"ɹ" },
	{ MONITOR_WRITE,	"write",	"˽" },
	{ MONITOR_FILL,		"fill",		"" },
	{ MONITOR_MOVE,		"move",		"ư" },
	{ MONITOR_SEARCH,	"search",	"򸡺" },
	{ MONITOR_OUT,		"out",		"ݡȽ" },
	{ MONITOR_LOADMEM,	"loadmem",	"ե뤫ɹ" },
	{ MONITOR_SAVEMEM,	"savemem",	"꤫ե˽" },
	{ MONITOR_RESET,	"reset",	"PC6001Vꥻå" },
	{ MONITOR_REG,		"reg",		"CPU쥸򻲾/" },
	{ MONITOR_DISASM,	"disasm",	"ե֥" },

};


////////////////////////////////////////////////////////////////
// μȽơ֥
////////////////////////////////////////////////////////////////
enum ArgvType{
	ARGV_END	= 0x00000,
	ARGV_STR	= 0x00001,	// strings
	ARGV_PORT	= 0x00002,	// 00xff
	ARGV_ADDR	= 0x00004,	// 00xffff
	ARGV_NUM	= 0x00008,	// 00x7fffffff
	ARGV_INT	= 0x00010,	// -0x7fffffff0x7fffffff
	ARGV_SIZE	= 0x00080,	// #1#0x7fffffff
	ARGV_REG	= 0x00400,	// RegisterName
	ARGV_BREAK	= 0x00800,	// BreakAction
	ARGV_STEP	= 0x04000,	// StepCommand
	
	EndofArgvType
};


enum ArgvName{
	// <reg>
	ARG_AF,		ARG_BC,		ARG_DE,		ARG_HL,
	ARG_IX,		ARG_IY,		ARG_SP,		ARG_PC,
	ARG_AF1,	ARG_BC1,	ARG_DE1,	ARG_HL1,
	ARG_I,		ARG_R,
	ARG_IFF,	ARG_IM,		ARG_HALT,
	
	// <action>
	//ARG_PC,
	ARG_READ,	ARG_WRITE,	ARG_IN,
	ARG_OUT,	ARG_CLEAR,
	
	// step <cmd>
	//ARG_ALL
	ARG_CALL,	ARG_JP,		ARG_REP,
	
	// reg all
	ARG_ALL,
	
	EndofArgName
};


struct{
	char *StrL;
	char *StrU;
	int Type;
	int Val;
}MonitorArgv[]=
{
	// <reg>
	{ "af",		"AF",		ARGV_REG,	ARG_AF,		},
	{ "bc",		"BC",		ARGV_REG,	ARG_BC,		},
	{ "de",		"DE",		ARGV_REG,	ARG_DE,		},
	{ "hl",		"HL",		ARGV_REG,	ARG_HL,		},
	{ "ix",		"IX",		ARGV_REG,	ARG_IX,		},
	{ "iy",		"IY",		ARGV_REG,	ARG_IY,		},
	{ "sp",		"SP",		ARGV_REG,	ARG_SP,		},
	{ "pc",		"PC",		ARGV_REG,	ARG_PC,		},
	{ "af'",	"AF'",		ARGV_REG,	ARG_AF1,	},
	{ "bc'",	"BC'",		ARGV_REG,	ARG_BC1,	},
	{ "de'",	"DE'",		ARGV_REG,	ARG_DE1,	},
	{ "hl'",	"HL'",		ARGV_REG,	ARG_HL1,	},
	{ "i",		"I",		ARGV_REG,	ARG_I,		},
	{ "r",		"R",		ARGV_REG,	ARG_R,		},
	{ "iff",	"IFF",		ARGV_REG,	ARG_IFF,	},
	{ "im",		"IM",		ARGV_REG,	ARG_IM,		},
	{ "halt",	"HALT",		ARGV_REG,	ARG_HALT,	},
	
	//<action>
	{ "pc",		"PC",		ARGV_BREAK,	ARG_PC,		},
	{ "read",	"READ",		ARGV_BREAK,	ARG_READ,	},
	{ "write",	"WRITE",	ARGV_BREAK,	ARG_WRITE,	},
	{ "in",		"IN",		ARGV_BREAK,	ARG_IN,		},
	{ "out",	"OUT",		ARGV_BREAK,	ARG_OUT,	},
	{ "clear",	"CLEAR",	ARGV_BREAK,	ARG_CLEAR,	},
	
	// step
	{ "call",	"CALL",		ARGV_STEP,	ARG_CALL,	},
	{ "jp",		"JP",		ARGV_STEP,	ARG_JP,		},
	{ "rep",	"REP",		ARGV_STEP,	ARG_REP,	},
	{ "all",	"ALL",		ARGV_STEP,	ARG_ALL,	},
};


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
cWndMon::cWndMon( VM6 *vm, const P6ID& id ) : P6DEVICE(vm,id)
{
	*KeyBuf = '\0';
	for( int i=0; i<MAX_HIS; i++ ) *(HisBuf[i]) = '\0';
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
cWndMon::~cWndMon( void ){}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
SDL_Surface *cWndMon::Init( void )
{
	SDL_Surface *surf = ZCons::Create( MONWINW, MONWINH, "" );
	
	if( surf ){
		// ǽåɽ
		ZCons::Printf( "***********************************************\n" );
		ZCons::Printf( "* PC6001V  - monitor mode -                   *\n" );
		ZCons::Printf( "*  help  ? Ϥȥإפɽޤ *\n" );
		ZCons::Printf( "***********************************************\n\n" );
		ZCons::Printf( PROMPT );
	}
	
	return surf;
}


////////////////////////////////////////////////////////////////
// ɥ
////////////////////////////////////////////////////////////////
void cWndMon::Update( void ){}


////////////////////////////////////////////////////////////////
// Ͻ
////////////////////////////////////////////////////////////////
void cWndMon::KeyIn( SDL_Event *event )
{
	static int LastKey  = SDLK_RETURN;	// Υ
	static int HisLevel = 1;			// ҥȥ٥
	
	if( event->type == SDL_KEYDOWN ){
		int KeyCode = event->key.keysym.sym;	// SDL Υ
		
		// ƥ󥭡б
		switch( KeyCode ){	// 
		case SDLK_KP0:			event->key.keysym.unicode = '0';	break;
		case SDLK_KP1:			event->key.keysym.unicode = '1';	break;
		case SDLK_KP2:			event->key.keysym.unicode = '2';	break;
		case SDLK_KP3:			event->key.keysym.unicode = '3';	break;
		case SDLK_KP4:			event->key.keysym.unicode = '4';	break;
		case SDLK_KP5:			event->key.keysym.unicode = '5';	break;
		case SDLK_KP6:			event->key.keysym.unicode = '6';	break;
		case SDLK_KP7:			event->key.keysym.unicode = '7';	break;
		case SDLK_KP8:			event->key.keysym.unicode = '8';	break;
		case SDLK_KP9:			event->key.keysym.unicode = '9';	break;
		case SDLK_KP_PERIOD:	event->key.keysym.unicode = '.';	break;
		case SDLK_KP_DIVIDE:	event->key.keysym.unicode = '/';	break;
		case SDLK_KP_MULTIPLY:	event->key.keysym.unicode = '*';	break;
		case SDLK_KP_MINUS:		event->key.keysym.unicode = '-';	break;
		case SDLK_KP_PLUS:		event->key.keysym.unicode = '+';	break;
		}
		
		
		switch( KeyCode ){	// 
		case SDLK_F6:		// ˥⡼ѹ
			vm->ToggleMonitor();
			break;
			
		case SDLK_RETURN:		// Enter
		case SDLK_KP_ENTER:		// Enter(ƥ󥭡)
			ZCons::Printf( "\n" );
			
			if( strlen( KeyBuf ) > 0 ){						// Хåեͭ
				if( stricmp( KeyBuf, HisBuf[1] ) != 0 ){	// ľΥҥȥȰۤʤʤ
					for( int i=MAX_HIS-1; i>1; i-- ) strcpy( HisBuf[i], HisBuf[i-1] );
					strcpy( HisBuf[1], KeyBuf );			// ХåեҥȥХåե˥ԡ
				}
			}
			
			// ǰϽ
			Exec( GetArg() );
			
			*KeyBuf = '\0';						// Хåեꥢ
			ZCons::Printf( PROMPT );
			break;
			
		case SDLK_UP:			// 
		case SDLK_DOWN:			// 
			if( LastKey == SDLK_UP || LastKey == SDLK_DOWN ){
				if( KeyCode == SDLK_UP   && HisLevel < MAX_HIS-1 && strlen( HisBuf[HisLevel+1] ) > 0 ) HisLevel++;
				if( KeyCode == SDLK_DOWN && HisLevel > 0 )                                             HisLevel--;
			}else{
				if( KeyCode == SDLK_UP ) HisLevel = 1;
				else                     HisLevel = 0;
			}
			
			// Υޥɥ饤õ
			while( strlen(KeyBuf) > 0 ){
				ZCons::LocateR( -1, 0 );
				ZCons::PutCharH( ' ' );
				ZCons::LocateR( -1, 0 );
				KeyBuf[strlen(KeyBuf)-1] = (Uint8)'\0';
			}
			// ҥȥХåե饭Хåե˥ԡɽ
			strcpy( KeyBuf, HisBuf[HisLevel] );
			ZCons::Printf( KeyBuf );
			break;
			
		case SDLK_BACKSPACE:	// BackSpace
			if( strlen(KeyBuf) > 0 ){
				ZCons::LocateR( -1, 0 );
				ZCons::PutCharH( ' ' );
				ZCons::LocateR( -1, 0 );
				KeyBuf[strlen(KeyBuf)-1] = (Uint8)'\0';
			}
			break;
			
		// ꥦɥ
		case SDLK_PAGEDOWN:	// PageDown
			vm->memw->SetAddress( vm->memw->GetAddress() - ( event->key.keysym.mod & KMOD_SHIFT ? 2048 : 16 ) );
			break;
			
		case SDLK_PAGEUP:	// PageUp
			vm->memw->SetAddress( vm->memw->GetAddress() + ( event->key.keysym.mod & KMOD_SHIFT ? 2048 : 16 ) );
			break;
			
		default:
			// ͭʸɤĥХåեդƤʤ
			if( ( event->key.keysym.unicode > 0x1f ) && ( event->key.keysym.unicode < 0x80 ) ){
				if( strlen(KeyBuf) < (MAX_CHRS-1) ){
					char *p = &KeyBuf[strlen(KeyBuf)];
					*p++ = (Uint8)event->key.keysym.unicode;
					*p   = (Uint8)'\0';
					ZCons::PutCharH( event->key.keysym.unicode );
				}
			}
		}
		
		LastKey = KeyCode;
	}
	
}


////////////////////////////////////////////////////////////////
// ֥졼ݥã
////////////////////////////////////////////////////////////////
void cWndMon::BreakIn( WORD addr )
{
	ZCons::SetColor( FC_YELLOW );
	ZCons::Printf( "\n << Break in %04XH >>", addr );
	switch( vm->bp->GetType( vm->bp->GetReqBPNum() ) ){
	case BPoint::BP_READ:	ZCons::Printf( " Read Memory %04XH",    vm->bp->GetAddr( vm->bp->GetReqBPNum() ) );	break;
	case BPoint::BP_WRITE:	ZCons::Printf( " Write Memory %04XH",   vm->bp->GetAddr( vm->bp->GetReqBPNum() ) );	break;
	case BPoint::BP_IN:		ZCons::Printf( " Read I/O Port %02XH",  vm->bp->GetAddr( vm->bp->GetReqBPNum() ) );	break;
	case BPoint::BP_OUT:	ZCons::Printf( " Write I/O Port %02XH", vm->bp->GetAddr( vm->bp->GetReqBPNum() ) );	break;
	default:				break;
	}
	ZCons::SetColor( FC_WHITE );
	ZCons::Printf( "\n" PROMPT );
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
int cWndMon::GetArg( void )
{
	char *p = KeyBuf;
	
	ArgvCounter = 1;	// Shift()ѥ󥿽
	Argc = 0;			// θĿ
	
	// ʬ
	while( (*p!='\0') && (Argc<=MAX_ARGS) ){
		if( *p==' ' )
			p++;
		else{
			Argv[Argc++] = p;
			while( (*p!='\0') && (*p!=' ') )
				p++;
			if( *p==' ' ) *p++ = '\0';
		}
	}
	
	int JobNo;
	if( !Argc )	// ?
		JobNo = MONITOR_NONE;
	else{		// ͭ̿?
		int i;
		for( i=0; i < COUNTOF(MonitorCmd); i++ )
				if( !strcmp( Argv[0], MonitorCmd[i].cmd ) ) break;
		
		if( i == COUNTOF( MonitorCmd ) ){	// ̵̿ξ
			ZCons::SetColor( FC_RED );
			ZCons::Printf( "̵ʥޥɤǤ : %s\n", Argv[0] );
			ZCons::SetColor( FC_WHITE );
			JobNo = MONITOR_NONE;
		}else{								//  ? ξ
			if( Argc==2 && !strcmp( Argv[1], "?" ) ){
				Help( MonitorCmd[i].Step );
				JobNo = MONITOR_NONE;
			}else{							// ̾̿ξ
				JobNo = MonitorCmd[i].Step;
				Shift();
			}
		}
	}
	
	return JobNo;
}


////////////////////////////////////////////////////////////////
// 󥷥ե
////////////////////////////////////////////////////////////////
void cWndMon::Shift( void )
{
	int i;
	BOOL size = FALSE;
	char *p, *chk;
	
	// ʾ̵
	if( ArgvCounter > MAX_ARGS || ArgvCounter >= Argc )
		argv.Type = ARGV_END;
	// ޤΤǲ
	else{
		p = Argv[ ArgvCounter ];
		if( *p == '#' ){ size = TRUE; p++; }
		
		argv.Type = 0;
		argv.Val  = strtol( p, &chk, 0 );
		argv.Str  = Argv[ ArgvCounter ];
		
		// ͤξ
		if( p!=chk && *chk=='\0' ){
			if( size ){		// #ǻϤޤ
				if( argv.Val <= 0 ) argv.Type = ARGV_STR;
				else                argv.Type = ARGV_SIZE;
			}else{			// ǻϤޤ
				argv.Type |= ARGV_INT;
				if( argv.Val >= 0 )      argv.Type |= ARGV_NUM;
				if( argv.Val <= 0xff )   argv.Type |= ARGV_PORT;
				if( argv.Val <= 0xffff ) argv.Type |= ARGV_ADDR;
			}
		// ʸξ
		}else{
			if( size ){		// #ǻϤޤ
				argv.Type = ARGV_STR;
			}else{			// ǻϤޤ
				for( i=0; i<COUNTOF( MonitorArgv ); i++ ){
					if( !strcmp( p, MonitorArgv[i].StrL ) || !strcmp( p, MonitorArgv[i].StrU ) ){
						argv.Type |= MonitorArgv[i].Type;
						argv.Val   = MonitorArgv[i].Val;
					}
				}
				if( !argv.Type ) argv.Type = ARGV_STR;
			}
		}
		ArgvCounter++;
	}
}


////////////////////////////////////////////////////////////////
// 顼å
////////////////////////////////////////////////////////////////
#define ErrorMes()													\
	do{																\
		ZCons::SetColor( FC_RED );									\
		ZCons::Printf( "̵Ǥ (arg %d)\n", ArgvCounter );	\
		ZCons::SetColor( FC_WHITE );								\
		return;														\
	}while(0)


////////////////////////////////////////////////////////////////
// ޥɼ¹
////////////////////////////////////////////////////////////////
void cWndMon::Exec( int cmd )
{
// 줿μå
#define	ArgvIs( x )	(argv.Type & (x))

	switch( cmd ){
	case MONITOR_HELP:
	//--------------------------------------------------------------
	// help [<cmd>]
	//	إפɽ
	//--------------------------------------------------------------
	{
		int i;
		char *cmd = NULL;
		
		if( argv.Type ){				// [cmd]
			cmd = argv.Str;
			Shift();
		}
		if( argv.Type ) ErrorMes();		// ;פʰХ顼
		
		if( !cmd ){	// ʤإɽ
			ZCons::Printf( "help\n" );
			for( i=0; i < COUNTOF(MonitorCmd); i++ )
				ZCons::Printf( "  %-7s %s\n", MonitorCmd[i].cmd, MonitorCmd[i].HelpMes );
			ZCons::Printf( "     : \"help <ޥ̾>\" Ϥ ˾ܺ٤ʥإפɽޤ\n" );
		}else{		// ΥޥɤΥإɽ
			for( i=0; i < COUNTOF(MonitorCmd); i++ )
				if( !strcmp( cmd, MonitorCmd[i].cmd ) ) break;
			if( i==COUNTOF(MonitorCmd) ) ErrorMes();
			Help( MonitorCmd[i].Step );
		}
		
		break;
	}
	case MONITOR_GO:
	//--------------------------------------------------------------
	//  go
	//	 ¹
	//--------------------------------------------------------------
	{
		if( argv.Type ) ErrorMes();
		
		vm->ToggleMonitor();
		
		break;
	}
	
	case MONITOR_TRACE:
	//--------------------------------------------------------------
	//  trace <step>
	//  trace #<step>
	//  ꤷƥåʬѤޤǼ¹
	//--------------------------------------------------------------
	{
		int step = 1;
		
		if( argv.Type ){
			if     ( ArgvIs( ARGV_SIZE ) ) step = argv.Val;	// [<step>]
			else if( ArgvIs( ARGV_NUM )  ) step = argv.Val;	// [#<step>]
			else                            ErrorMes();
			Shift();
		}
		
		if( argv.Type ) ErrorMes();
		
		
		while( step-- ){
			vm->sche->Update( vm->cpu->Exec() );	// 1̿¹Ԥȥ٥ȹ
		}
		
//		if( CheckBreakPointPC() ) set_emumode( TRACE_BP );
//		else                      set_emumode( M_TRACE );
		
		break;
	}
	
	case MONITOR_STEP:
	//--------------------------------------------------------------
	//  step
	//  step [call] [jp] [rep] [all]
	//  1ƥå,¹
	//  CALL,DJNZ,LDIR etc Υåפ
	//--------------------------------------------------------------
	{
		BOOL call = FALSE, jp = FALSE, rep = FALSE;
		BOOL flag = FALSE;
		BYTE code;
		WORD addr;
		char DisCode[128];
		Register reg;
		
		while( argv.Type ){
			if( ArgvIs( ARGV_STEP ) ){
				if( argv.Val == ARG_CALL )	call = TRUE;
				if( argv.Val == ARG_JP )	jp   = TRUE;
				if( argv.Val == ARG_REP )	rep  = TRUE;
				if( argv.Val == ARG_ALL )	call = jp = rep = TRUE;
				Shift();
			}else
				ErrorMes();
		}
		
		if( argv.Type ) ErrorMes();
		
		vm->cpu->GetRegister( &reg );
		
		addr = reg.PC.W;
		code = vm->mem->Read( addr );
		
		if( call ){
			if( code		== 0xcd ||	// CALL nn    = 11001101B
			  ( code&0xc7 ) == 0xc4 ){	// CALL cc,nn = 11ccc100B
				addr += 3;
				flag = TRUE;
			}
		}
		
		if( jp ){
			if( code == 0x10 ){			// DJNZ e     = 00010000B
				addr += 2;
				flag = TRUE;
			}
		}
		
	    if( rep ){
			if( code == 0xed ){			// LDIR/LDDR/CPIR/CPDR etc
				code = vm->mem->Read( addr+1 );
				if( (code&0xf4) == 0xb0 ){
					addr += 2;
					flag = TRUE;
				}
			}
		}
		
		vm->cpu->Disasm( DisCode, addr );
		ZCons::Printf( "%s\n", DisCode );
		
		vm->sche->Update( vm->cpu->Exec() );	// 1̿¹Ԥȥ٥ȹ
		
//		if( flag ){
//			BP[BP_NUM_FOR_SYSTEM].Type = BP_PC;
//			BP[BP_NUM_FOR_SYSTEM].Addr   = addr;
//			set_emumode( STEP_BP );
//		}else{
//			BP[BP_NUM_FOR_SYSTEM].Type = BP_NONE;
//			set_emumode( STEP );
//		}
		
		
		break;
	}	
	case MONITOR_STEPALL:
	//--------------------------------------------------------------
	//  S
	//  step all Ʊ
	//--------------------------------------------------------------
	{
		BOOL flag = FALSE;
		BYTE code;
		WORD addr;
		char DisCode[128];
		Register reg;
		
		if( argv.Type ) ErrorMes();
		
		vm->cpu->GetRegister( &reg );
		
		addr = reg.PC.W;
		code = vm->mem->Read( addr );
		
		if( code		== 0xcd ||	// CALL nn    = 11001101B
		  ( code&0xc7 ) == 0xc4 ){	// CALL cc,nn = 11ccc100B
			addr += 3;
			flag = TRUE;
		}
		
		if( code == 0x10 ){			// DJNZ e     = 00010000B
			addr += 2;
			flag = TRUE;
		}
		
		if( code == 0xed ){			// LDIR/LDDR/CPIR/CPDR etc
			code = vm->mem->Read( addr+1 );
			if( (code&0xf4) == 0xb0 ){
				addr += 2;
				flag = TRUE;
			}
		}
		
		vm->cpu->Disasm( DisCode, addr );
		ZCons::Printf( "%s\n", DisCode );
		
		vm->sche->Update( vm->cpu->Exec() );	// 1̿¹Ԥȥ٥ȹ
		
//		if( flag ){
//			BP[BP_NUM_FOR_SYSTEM].Type = BP_PC;
//			BP[BP_NUM_FOR_SYSTEM].Addr   = addr;
//			set_emumode( STEP_BP );
//		}else{
//			BP[BP_NUM_FOR_SYSTEM].Type = BP_NONE;
//			set_emumode( STEP );
//		}
		
		break;
	}	
	case MONITOR_BREAK:
	//--------------------------------------------------------------
	//  break [PC|READ|WRITE|IN|OUT] <addr|port> [#<No>]
	//  break CLEAR [#<No>]
	//  break
	//  ֥졼ݥȤ꡿ɽ
	//--------------------------------------------------------------
	{
		BOOL show = FALSE;
		int action = ARG_PC;
		WORD addr = 0;
		int number = 0;
		
		if( argv.Type ){
			// <action>
			if( ArgvIs( ARGV_BREAK ) ){
				action = argv.Val;
				Shift();
			}
			
			// <addr|port>
			switch( action ){
			case ARG_IN:
			case ARG_OUT:
				if( !ArgvIs( ARGV_PORT ) ) ErrorMes();
				addr = argv.Val;
				Shift();
				break;
			case ARG_PC:
			case ARG_READ:
			case ARG_WRITE:
				if( !ArgvIs( ARGV_ADDR ) ) ErrorMes();
				addr = argv.Val;
				Shift();
				break;
			}
			
			// [#<No>]
			if( argv.Type ){
				if( !ArgvIs( ARGV_SIZE ) ) ErrorMes();
				if( argv.Val < 1 || argv.Val > NR_BP ) ErrorMes();
				number = argv.Val - 1;
				Shift();
			}
		}else{
			show = TRUE;
		}
		
		if( argv.Type ) ErrorMes();
		
		
		if( show ){
			for( int i=0; i<NR_BP; i++ ){
				ZCons::Printf( "    #%02d  ", i+1 );
				addr = vm->bp->GetAddr( i );
				switch( vm->bp->GetType( i ) ){
				case BPoint::BP_NONE:
					ZCons::Printf( "-- ʤ --\n" );
					break;
				case BPoint::BP_PC:
					ZCons::Printf( "PC   reach %04XH\n", addr&0xffff );
					break;
				case BPoint::BP_READ:
					ZCons::Printf( "READ  from %04XH\n", addr&0xffff );
					break;
				case BPoint::BP_WRITE:
					ZCons::Printf( "WRITE   to %04XH\n", addr&0xffff );
					break;
				case BPoint::BP_IN:
					ZCons::Printf( "INPUT from %02XH\n", addr&0xff );
					break;
				case BPoint::BP_OUT:
					ZCons::Printf( "OUTPUT  to %04XH\n", addr&0xff) ;
					break;
				default:
					break;
				}
			}
		}else{
			if( action == ARG_CLEAR ){
				vm->bp->SetType( number, BPoint::BP_NONE );
				ZCons::Printf( "֥졼ݥ #%02d õޤ\n", number+1 );
			}else{
				char *s = NULL;
				
				switch( action ){
				case ARG_PC:
					vm->bp->SetType( number, BPoint::BP_PC );
					s = "PC : %04XH";
					break;
				case ARG_READ:
					vm->bp->SetType( number, BPoint::BP_READ );
					s = "READ : %04XH";
					break;
				case ARG_WRITE:
					vm->bp->SetType( number, BPoint::BP_WRITE );
					s = "WRITE : %04XH";
					break;
				case ARG_IN:
					vm->bp->SetType( number, BPoint::BP_IN );
					s = "IN : %02XH";
					break;
				case ARG_OUT:
					vm->bp->SetType( number, BPoint::BP_OUT );
					s = "OUT : %02XH";
					break;
				}
				vm->bp->SetAddr( number, addr );
				ZCons::Printf( "֥졼ݥ #%02d ꤷޤ[ ",number+1 );
				ZCons::Printf( s, addr );
				ZCons::Printf( " ]\n" );
			}
		}
		break;
	}
	case MONITOR_READ:
	//--------------------------------------------------------------
	//  read <addr>
	//  Υɥ쥹꡼
	//--------------------------------------------------------------
		break;
		
	case MONITOR_WRITE:
	//--------------------------------------------------------------
	//  write <addr> <data>
	//  Υɥ쥹˥饤
	//--------------------------------------------------------------
	{
		if( !argv.Type ) ErrorMes();
		
		// <addr>
		if( !ArgvIs( ARGV_ADDR )) ErrorMes();
		WORD addr = argv.Val;
		Shift();
		
		// <data>
		if( !ArgvIs( ARGV_INT )) ErrorMes();
		BYTE data = argv.Val;
		Shift();
		
		if( argv.Type ) ErrorMes();
		
		vm->mem->Write( addr, data );
		
		ZCons::Printf( "WRITE memory [ %04XH ] <- %02X  (= %d | %+d | ", addr, (Uint8)data, (Uint8)data, (Sint8)data );
		int i,j;
		for( i=0, j=0x80; i<8; i++, j>>=1 ) ZCons::Printf( "%d", (data & j) ? 1 : 0 );
		ZCons::Printf( "B )\n");
		
		break;
	}
	
	case MONITOR_FILL:
	//--------------------------------------------------------------
	//  fill <start-addr> <end-addr> <value>
	//  fill <start-addr> #<size> <value>
	//  
	//--------------------------------------------------------------
	{
		int start, size, value;
		
		if( !argv.Type ) ErrorMes();
		
		// <addr>
		if( !ArgvIs( ARGV_ADDR ) ) ErrorMes();
		start = argv.Val;
		Shift();						
		
		// [<addr|#size>]
		if     ( ArgvIs( ARGV_SIZE ) ) size = argv.Val;
		else if( ArgvIs( ARGV_ADDR ) ) size = argv.Val - start +1;
		else                           ErrorMes();
		Shift();
		
		// <data>
		if( !ArgvIs( ARGV_INT )) ErrorMes();
		value = argv.Val;
		Shift();
		
		if( argv.Type ) ErrorMes();
		
		for( int i=0; i<size; i++ ) vm->mem->Write( start+i, value );
		
		break;
	}
	
	case MONITOR_MOVE:
	//--------------------------------------------------------------
	//  move <src-addr> <end-addr> <dist-addr>
	//  move <src-addr> #size      <dist-addr>
	//  ž
	//--------------------------------------------------------------
	{
		int start, size, dist;
		
		if( !argv.Type ) ErrorMes();
		
		// <addr>
		if( !ArgvIs( ARGV_ADDR ) ) ErrorMes();
		start = argv.Val;
		Shift();
		
		// [<addr|#size>]
		if     ( ArgvIs( ARGV_SIZE ) ) size = argv.Val;
		else if( ArgvIs( ARGV_ADDR ) ) size = argv.Val - start +1;
		else                           ErrorMes();
		Shift();
		
		// <addr>
		if( !ArgvIs( ARGV_ADDR ) ) ErrorMes();
		dist = argv.Val;
		Shift();
		
		if( argv.Type ) ErrorMes();
		
		// ž-ž褬 Ťʤʤ
		if( start+size <= dist ) for( int i=0; i<size; i++ )    vm->mem->Write( dist+i, vm->mem->Read( start+i ) );
		// ž-ž褬 Ťʤ
		else                     for( int i=size-1; i>=0; i-- ) vm->mem->Write( dist+i, vm->mem->Read( start+i ) );
		
		break;
	}
	
	case MONITOR_SEARCH:
	//--------------------------------------------------------------
	//  search [<value> [<start-addr> <end-addr>]]
	//  search [<value> [<start-addr> #<size>]]
	//   (1Х) 򥵡
	//--------------------------------------------------------------
		break;
		
	case MONITOR_OUT:
	//--------------------------------------------------------------
	//  out <port> <data>
	//  ΥݡȤ˽
	//--------------------------------------------------------------
		break;
		
	case MONITOR_LOADMEM:
	//--------------------------------------------------------------
	// loadmem <filename> <start-addr> <end-addr>
	// loadmem <filename> <start-addr> #<size>
	// ե뤫˥
	//--------------------------------------------------------------
		break;
		
	case MONITOR_SAVEMEM:
	//--------------------------------------------------------------
	//  savemem <filename> <start-addr> <end-addr>
	//  savemem <filename> <start-addr> #<size>
	//  ե˥
	//--------------------------------------------------------------
		break;
		
	case MONITOR_RESET:
	//--------------------------------------------------------------
	//  reset
	//	ꥻå
	//--------------------------------------------------------------
		if( argv.Type ) ErrorMes();
		vm->Reset();
		
		break;
		
	case MONITOR_REG:
	//--------------------------------------------------------------
	//  reg <name> <value>
	//  쥸Ƥѹ
	//--------------------------------------------------------------
	{	int re = -1, val=0, i;
		char *str;
		Register reg;
		
		if( argv.Type ){
			if( !ArgvIs( ARGV_REG )) ErrorMes();		// <name>
			re = argv.Val;
			Shift();
			if( !ArgvIs( ARGV_INT )) ErrorMes();		// <value>
			val = argv.Val;
			Shift();
		}
		if( argv.Type ) ErrorMes();
		
		vm->cpu->GetRegister( &reg );
		
		switch( re ){
		case ARG_AF:	reg.AF.W = val;		break;
		case ARG_BC:	reg.BC.W = val;		break;
		case ARG_DE:	reg.DE.W = val;		break;
		case ARG_HL:	reg.HL.W = val;		break;
		case ARG_IX:	reg.IX.W = val;		break;
		case ARG_IY:	reg.IY.W = val;		break;
		case ARG_SP:	reg.SP.W = val;		break;
		case ARG_PC:	reg.PC.W = val;		break;
		case ARG_AF1:	reg.AF1.W = val;	break;
		case ARG_BC1:	reg.BC1.W = val;	break;
		case ARG_DE1:	reg.DE1.W = val;	break;
		case ARG_HL1:	reg.HL1.W = val;	break;
		case ARG_I:		val &= 0xff; reg.I = val;		break;
		case ARG_R:		val &= 0xff; reg.R = val;		break;
		case ARG_IFF:	if(val)   val=1; reg.IFF  = val;	break;
		case ARG_IM:	if(val>3) val=2; reg.IM   = val;	break;
		case ARG_HALT:	if(val)   val=1; reg.Halt = val;	break;
		}
		
		vm->cpu->SetRegister( &reg );
		
		for( i=0; i<COUNTOF( MonitorArgv ); i++ )
			if( re == MonitorArgv[i].Val ) break;
		if( i == COUNTOF( MonitorArgv ) ) str = "";
		else                              str = MonitorArgv[i].StrU;
		ZCons::Printf( "reg %s <- %04X\n", str, val );
		
		break;
	}
		
	case MONITOR_DISASM:
	//--------------------------------------------------------------
	//  disasm [[<start-addr>][#<steps>]]
	//  ե֥
	//--------------------------------------------------------------
	{	static int SaveDisasmAddr = -1;
		int i, pc;
		int addr = SaveDisasmAddr;
		int step = 16;
		char DisCode[128];
		Register reg;
		
		if( argv.Type ){
			if( ArgvIs( ARGV_ADDR )){		// [<addr>]
				addr = argv.Val;
				Shift();
			}
			if( ArgvIs( ARGV_SIZE )){		// [#<step>]
				step = argv.Val;
				Shift();
			}
		}
		if( argv.Type ) ErrorMes();
		
		vm->cpu->GetRegister( &reg );
		if( addr == -1 ) addr = reg.PC.W;	// ADDR ̤
		
		pc = 0;
		for( i=0; i<step; i++ ){
			pc += vm->cpu->Disasm( DisCode, (WORD)(addr+pc) );
			ZCons::Printf( "%s\n", DisCode );
		}
		SaveDisasmAddr = ( addr + pc ) & 0xffff;
		break;
	}
	
	}
}


////////////////////////////////////////////////////////////////
// إɽ
////////////////////////////////////////////////////////////////
void cWndMon::Help( int cmd )
{
	switch( cmd ){
	case MONITOR_HELP:
		ZCons::Printf(
			"  help [<cmd>]\n"
			"    إפɽޤ\n"
			"    <cmd> ... إפɽޥ\n"
			"              [omit]... ޥɤδñʥإפɽޤ\n"
		);
		break;
		
	case MONITOR_GO:
		ZCons::Printf(
			"  go\n"
			"    ץ¹Ԥޤ\n"
		);
		break;
		
	case MONITOR_TRACE:
		ZCons::Printf(
			"  trace [#<steps>|<steps>]\n"
			"    execute program specityes times\n"
			"    [all omit]        ... trace some steps (previous steps)\n"
			"    #<steps>, <steps> ... step counts of trace  ( you can omit '#' )\n"
		);
		break;
		
	case MONITOR_STEP:
		ZCons::Printf(
			"  step [call][jp][rep]\n"
			"    execute program 1 time\n"
			"    [all omit] ... execute 1 step\n"
			"    call       ... not trace CALL instruction\n"
			"    jp         ... not trace DJNZ instruction\n"
			"    rep        ... not trace LD*R/CP*R/IN*R/OT*R instruction\n"
			"    CAUTION)\n"
			"         call/jp/rep are use break-point #10.\n"
		);
		break;
		
	case MONITOR_STEPALL:
		ZCons::Printf(
			"  S\n"
			"    'step all' Ʊ   (step򻲾)\n"
		);
		break;
		
	case MONITOR_BREAK:
		ZCons::Printf(
			"  break [<action>] <addr|port> [#<No>]\n"
			"  break CLEAR [#<No>]\n"
			"  break\n"
			"    ֥졼ݥȤꤷޤ\n"
			"    [all omit]  ... ƤΥ֥졼ݥȤɽ\n"
			"    <action>    ... set action of conditon PC|READ|WRITE|IN|OUT or CLEAR\n"
			"                    PC    ... break if PC reach addr\n"
			"                    READ  ... break if data is read\n"
			"                    WRITE ... break if data is written\n"
			"                    IN    ... break if data is input\n"
			"                    OUT   ... break if data is output\n"
			"                    CLEAR ... clear all break point\n"
			"                    [omit]... select PC\n"
			"    <addr|port> ... specify address or port\n"
			"                    if <action> is CLEAR, this argument is invalid\n"
			"    #<No>       ... number of break point. (#1..#10)\n"
			"                    [omit]... select #1\n"
			"                    CAUTION).. #10 is used by system\n"
		);
		break;
		
	case MONITOR_READ:
		ZCons::Printf(
			"  read <addr>\n"
			"    ɹߤޤ\n"
			"    <addr> ... ꥢɥ쥹\n"
		);
		break;
		
	case MONITOR_WRITE:
		ZCons::Printf(
			"  write <addr> <data>\n"
			"    ˽ߤޤ\n"
			"    <addr> ... ꥢɥ쥹\n"
			"    <data> ... ǡ\n"
		);
		break;
		
	case MONITOR_FILL:
		ZCons::Printf(
			"  fill <start-addr> <end-addr> <value>\n"
			"  fill <start-addr> #<size>    <value>\n"
			"    ͤޤ\n"
			"    <start-addr> ... ϥɥ쥹\n"
			"    <end-addr>   ... λɥ쥹\n"
			"    #<size>      ... \n"
			"    <value>      ... \n"
		);
		break;
		
	case MONITOR_MOVE:
		ZCons::Printf(
			"  move <src-addr> <end-addr> <dist-addr>\n"
			"  move <src-addr> #<size>    <dist-addr>\n"
			"    žޤ\n"
			"    <src-addr>  ... žϥɥ쥹\n"
			"    <end-addr>  ... žλɥ쥹\n"
			"    #<size>     ... ž\n"
			"    <dist-addr> ... ž襢ɥ쥹\n"
		);
		break;
		
	case MONITOR_SEARCH:
		ZCons::Printf(
			"  search [<value> [<start-addr> <end-addr>]]\n"
			"  search [<value> [<start-addr> #<size>]]\n"
			"    򸡺ޤ\n"
			"    <value>      ... \n"
			"    <start-addr> ... ϥɥ쥹\n"
			"    <end-addr>   ... λɥ쥹\n"
			"    #<size>      ... \n"
			"    [omit-all]   ... ͤޤʸ򸡺\n"
		);
		break;
		
	case MONITOR_OUT:
		ZCons::Printf(
			"  out <port> <data>\n"
			"    I/OݡȤ˽Ϥޤ\n"
			"    <port> ... I/Oݡȥɥ쥹\n"
			"    <data> ... ϥǡ\n"
		);
		break;
		
	case MONITOR_LOADMEM:
		ZCons::Printf(
			"  loadmem <filename> <start-addr> <end-addr>\n"
			"  loadmem <filename> <start-addr> #<size>\n"
			"    ե뤫˥ɤޤ\n"
			"    <filename>   ... ե̾\n"
			"    <start-addr> ... ɳϥɥ쥹\n"
			"    <end-addr>   ... ɽλɥ쥹\n"
			"    #<size>      ... ɥ\n"
		);
		break;
		
	case MONITOR_SAVEMEM:
		ZCons::Printf(
			"  savemem <filename> <start-addr> <end-addr>\n"
			"  savemem <filename> <start-addr> #<size>\n"
			"    ꥤ᡼ե˥֤ޤ\n"
			"    <filename>   ... ե̾\n"
			"    <start-addr> ... ֳϥɥ쥹\n"
			"    <end-addr>   ... ֽλɥ쥹\n"
			"    #<size>      ... ֥\n"
		);
		break;
		
	case MONITOR_RESET:
		ZCons::Printf(
			"  reset\n"
			"    PC6001VꥻåȤɥ쥹 0000H ¹Ԥޤ\n"
		);
		break;
		
	case MONITOR_REG:
		ZCons::Printf(
			"  reg <name> <value>\n"
			"    쥸ͤ򻲾ȡꤷޤ\n"
			"    <name>     ... specity register name.\n"
			"                   AF|BC|DE|HL|AF'|BC'|DE'|HL'|IX|IY|SP|PC|I|R|IFF|IM\n"
			"    <value>    ... set value\n"
		);
		break;
		
	case MONITOR_DISASM:
		ZCons::Printf(
			"  disasm [[<start-addr>][#<steps>]]\n"
			"    ե֥뤷ޤ\n"
			"    [all omit]   ... PC쥸ɥ쥹16ƥåʬե֥뤷ޤ\n"
			"    <start-addr> ... start-addr ե֥뤷ޤ\n"
			"                     [omit]... PC쥸ɥ쥹\n"
			"    #<steps>     ... ե֥뤹륹ƥå׿\n"
			"                     [omit]... 16ƥå\n"
		);
		break;
		
	}
}
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
