#include <string.h>
#include <memory.h>
#include <stdlib.h>

#include "keyboard.h"
#include "common.h"
#include "joystick.h"
#include "osd.h"
#include "pd8049.h"
#include "schedule.h"
#include "tape.h"


// ٥ID
#define	EID_AKEY	(1)

// 
#define	WAIT_AKEY	(10)	/* ms */


// SDL -> ۥ
static const struct {
	int InKey;			// SDLΥ
	PCKEYsym VKey;		// ۥ
} VKeyDef[] = {
	{ SDLK_UNKNOWN,			KVC_UNKNOWN },
	
	{ SDLK_1,				KVC_1 },			// 1	!
	{ SDLK_2,				KVC_2 },			// 2	"
	{ SDLK_3,				KVC_3 },			// 3	#
	{ SDLK_4,				KVC_4 },			// 4	$
	{ SDLK_5,				KVC_5 },			// 5	%
	{ SDLK_6,				KVC_6 },			// 6	&
	{ SDLK_7,				KVC_7 },			// 7	'
	{ SDLK_8,				KVC_8 },			// 8	(
	{ SDLK_9,				KVC_9 },			// 9	)
	{ SDLK_0,				KVC_0 },			// 0
	
	{ SDLK_a,				KVC_A },			// a	A
	{ SDLK_b,				KVC_B },			// b	B
	{ SDLK_c,				KVC_C },			// c	C
	{ SDLK_d,				KVC_D },			// d	D
	{ SDLK_e,				KVC_E },			// e	E
	{ SDLK_f,				KVC_F },			// f	F
	{ SDLK_g,				KVC_G },			// g	G
	{ SDLK_h,				KVC_H },			// h	H
	{ SDLK_i,				KVC_I },			// i	I
	{ SDLK_j,				KVC_J },			// j	J
	{ SDLK_k,				KVC_K },			// k	K
	{ SDLK_l,				KVC_L },			// l	L
	{ SDLK_m,				KVC_M },			// m	M
	{ SDLK_n,				KVC_N },			// n	N
	{ SDLK_o,				KVC_O },			// o	O
	{ SDLK_p,				KVC_P },			// p	P
	{ SDLK_q,				KVC_Q },			// q	Q
	{ SDLK_r,				KVC_R },			// r	R
	{ SDLK_s,				KVC_S },			// s	S
	{ SDLK_t,				KVC_T },			// t	T
	{ SDLK_u,				KVC_U },			// u	U
	{ SDLK_v,				KVC_V },			// v	V
	{ SDLK_w,				KVC_W },			// w	W
	{ SDLK_x,				KVC_X },			// x	X
	{ SDLK_y,				KVC_Y },			// y	Y
	{ SDLK_z,				KVC_Z },			// z	Z
	
	{ SDLK_F1,				KVC_F1 },			// F1
	{ SDLK_F2,				KVC_F2 },			// F2
	{ SDLK_F3,				KVC_F3 },			// F3
	{ SDLK_F4,				KVC_F4 },			// F4
	{ SDLK_F5,				KVC_F5 },			// F5
	{ SDLK_F6,				KVC_F6 },			// F6
	{ SDLK_F7,				KVC_F7 },			// F7
	{ SDLK_F8,				KVC_F8 },			// F8
	{ SDLK_F9,				KVC_F9 },			// F9
	{ SDLK_F10,				KVC_F10 },			// F10
	{ SDLK_F11,				KVC_F11 },			// F11
	{ SDLK_F12,				KVC_F12 },			// F12
	
	{ SDLK_MINUS,			KVC_MINUS },		// -	=
	{ SDLK_CARET,			KVC_CARET },		// ^	~
	{ SDLK_BACKSPACE,		KVC_BACKSPACE },	// BackSpace
	{ SDLK_AT,				KVC_AT },			// @	`
	{ SDLK_LEFTBRACKET,		KVC_LBRACKET },		// [	{
	{ SDLK_SEMICOLON,		KVC_SEMICOLON },	// ;	+
	{ SDLK_COLON,			KVC_COLON },		// :	*
	{ SDLK_COMMA,			KVC_COMMA },		// ,	<
	{ SDLK_PERIOD,			KVC_PERIOD },		// .	>
	{ SDLK_SLASH,			KVC_SLASH },		// /	?
	{ SDLK_SPACE,			KVC_SPACE },		// Space
	
	{ SDLK_ESCAPE,			KVC_ESC },			// ESC
	{ SDLK_BACKQUOTE,		KVC_HANZEN },		// Ⱦ/
	{ SDLK_TAB,				KVC_TAB },			// Tab
	{ SDLK_CAPSLOCK,		KVC_CAPSLOCK },		// CapsLock
	{ SDLK_RETURN,			KVC_ENTER },		// Enter
	{ SDLK_LCTRL,			KVC_LCTRL },		// L-Ctrl
	{ SDLK_RCTRL,			KVC_RCTRL },		// R-Ctrl
	{ SDLK_LSHIFT,			KVC_LSHIFT },		// L-Shift
	{ SDLK_RSHIFT,			KVC_RSHIFT },		// R-Shift
	{ SDLK_LALT,			KVC_LALT },			// L-Alt
	{ SDLK_RALT,			KVC_RALT },			// R-Alt
	{ SDLK_PRINT,			KVC_PRINT },		// PrintScreen
	{ SDLK_SCROLLOCK,		KVC_SCROLLLOCK },	// ScrollLock
	{ SDLK_PAUSE,			KVC_PAUSE },		// Pause
	{ SDLK_INSERT,			KVC_INSERT },		// Insert
	{ SDLK_DELETE,			KVC_DELETE },		// Delete
	{ SDLK_END,				KVC_END },			// End
	{ SDLK_HOME,			KVC_HOME },			// Home
	{ SDLK_PAGEUP,			KVC_PAGEUP },		// PageUp
	{ SDLK_PAGEDOWN,		KVC_PAGEDOWN },		// PageDown
	
	{ SDLK_UP,				KVC_UP },			// 
	{ SDLK_DOWN,			KVC_DOWN },			// 
	{ SDLK_LEFT,			KVC_LEFT },			// 
	{ SDLK_RIGHT,			KVC_RIGHT },		// 
	
	{ SDLK_KP0,				KVC_P0 },			// [0]
	{ SDLK_KP1,				KVC_P1 },			// [1]
	{ SDLK_KP2,				KVC_P2 },			// [2]
	{ SDLK_KP3,				KVC_P3 },			// [3]
	{ SDLK_KP4,				KVC_P4 },			// [4]
	{ SDLK_KP5,				KVC_P5 },			// [5]
	{ SDLK_KP6,				KVC_P6 },			// [6]
	{ SDLK_KP7,				KVC_P7 },			// [7]
	{ SDLK_KP8,				KVC_P8 },			// [8]
	{ SDLK_KP9,				KVC_P9 },			// [9]
	{ SDLK_NUMLOCK,			KVC_NUMLOCK },		// NumLock
	{ SDLK_KP_PLUS,			KVC_P_PLUS },		// [+]
	{ SDLK_KP_MINUS,		KVC_P_MINUS },		// [-]
	{ SDLK_KP_MULTIPLY,		KVC_P_MULTIPLY },	// [*]
	{ SDLK_KP_DIVIDE,		KVC_P_DIVIDE },		// [/]
	{ SDLK_KP_PERIOD,		KVC_P_PERIOD },		// [.]
	{ SDLK_KP_ENTER,		KVC_P_ENTER },		// [Enter]
	
	// ܸ쥭ܡɤΤ
	{ SDLK_BACKSLASH,		KVC_YEN },			// 	|
	{ SDLK_RIGHTBRACKET,	KVC_RBRACKET },		// ]	}
	{ SDLK_UNDERSCORE,		KVC_UNDERSCORE },	// 	_
//	{ 				,		KVC_MUHENKAN },		// ̵Ѵ
//	{ 				,		KVC_HENKAN },		// Ѵ
//	{ 				,		KVC_HIRAGANA },		// Ҥ餬
	
	// Ѹ쥭ܡɤΤ
	{ SDLK_BACKSLASH,		KVE_BACKSLASH },	// 	|
	
	// ɲå
	{ SDLK_LMETA,			KVX_LMETA },		// L-Meta
	{ SDLK_RMETA,			KVX_RMETA },		// R-Meta
	{ SDLK_MENU,			KVX_MENU }			// Menu
};


// P6 -> P6ޥȥꥯ
static const struct {
	P6KEYsym P6KEY;		// P6
	BYTE Mat;			// P6Υޥȥꥯ(bit7-4:Y bit3-0:X)
} P6KeyMx[] = {
	
	{ KP6_UNKNOWN,		0x00 },	// 
	
	// Y0
	{ KP6_CTRL,			0x01 },	// CTRL
	{ KP6_SHIFT,		0x02 },	// SHIFT
	{ KP6_GRAPH,		0x03 },	// GRAPH
	
	// Y1
	{ KP6_1,			0x10 },	// 1	!
	{ KP6_Q,			0x11 },	// q	Q
	{ KP6_A,			0x12 },	// a	A
	{ KP6_Z,			0x13 },	// z	Z
	{ KP6_K,			0x14 },	// k	K
	{ KP6_I,			0x15 },	// i	I
	{ KP6_8,			0x16 },	// 8	(
	{ KP6_COMMA,		0x17 },	// ,	<
	
	// Y2
	{ KP6_2,			0x20 },	// 2	"
	{ KP6_W,			0x21 },	// w	W
	{ KP6_S,			0x22 },	// s	S
	{ KP6_X,			0x23 },	// x	X
	{ KP6_L,			0x24 },	// l	L
	{ KP6_O,			0x25 },	// o	O
	{ KP6_9,			0x26 },	// 9	)
	{ KP6_PERIOD,		0x27 },	// .	>
	
	// Y3
	{ KP6_3,			0x30 },	// 3	#
	{ KP6_E,			0x31 },	// e	E
	{ KP6_D,			0x32 },	// d	D
	{ KP6_C,			0x33 },	// c	C
	{ KP6_SEMICOLON,	0x34 },	// ;	+
	{ KP6_P,			0x35 },	// p	P
	{ KP6_F1,			0x36 },	// F1	F6
	{ KP6_SLASH,		0x37 },	// /	?
	
	// Y4
	{ KP6_4,			0x40 },	// 4	$
	{ KP6_R,			0x41 },	// r	R
	{ KP6_F,			0x42 },	// f	F
	{ KP6_V,			0x43 },	// v	V
	{ KP6_COLON,		0x44 },	// :	*
	{ KP6_AT,			0x45 },	// @
	{ KP6_F2,			0x46 },	// F2	F7
	{ KP6_UNDERSCORE,	0x47 },	// 		_
	
	// Y5
	{ KP6_5,			0x50 },	// 5	%
	{ KP6_T,			0x51 },	// t	T
	{ KP6_G,			0x52 },	// g	G
	{ KP6_B,			0x53 },	// b	B
	{ KP6_RBRACKET,		0x54 },	// ]
	{ KP6_LBRACKET,		0x55 },	// [
	{ KP6_F3,			0x56 },	// F3	F8
	{ KP6_SPACE,		0x57 },	// SPC
	
	// Y6
	{ KP6_6,			0x60 },	// 6	&
	{ KP6_Y,			0x61 },	// y	Y
	{ KP6_H,			0x62 },	// h	H
	{ KP6_N,			0x63 },	// n	N
	{ KP6_MINUS,		0x64 },	// -	=
	{ KP6_CARET,		0x65 },	// ^
	{ KP6_F4,			0x66 },	// F4	F9
	{ KP6_0,			0x67 },	// 0
	
	// Y7
	{ KP6_7,			0x70 },	// 7	'
	{ KP6_U,			0x71 },	// u	U
	{ KP6_J,			0x72 },	// j	J
	{ KP6_M,			0x73 },	// m	M
	{ KP6_YEN,			0x75 },	// 
	{ KP6_F5,			0x76 },	// F5	F10
	
	// Y8
	{ KP6_RETURN,		0x80 },	// RET
	{ KP6_STOP,			0x81 },	// STOP
	{ KP6_UP,			0x82 },	// 
	{ KP6_DOWN,			0x83 },	// 
	{ KP6_RIGHT,		0x84 },	// 
	{ KP6_LEFT,			0x85 },	// 
	{ KP6_TAB,			0x86 },	// TAB
	{ KP6_ESC,			0x87 },	// ESC
	
	// Y9
	{ KP6_KANA,			0x90 },	// 
	{ KP6_INS,			0x91 },	// INS
	{ KP6_DEL,			0x92 },	// DEL
	{ KP6_PAGE,			0x93 },	// PAGE
	{ KP6_HOME,			0x94 },	// HOME
	{ KP6_MODE,			0x95 },	// MODE
	{ KP6_CAPS,			0x96 },	// CAPS
	
	// ʹߤȼĥ
	// YA
	{ KP6_P0,			0xa0 },	// [0]
	{ KP6_P1,			0xa1 },	// [1]
	{ KP6_P2,			0xa2 },	// [2]
	{ KP6_P3,			0xa3 },	// [3]
	{ KP6_P4,			0xa4 },	// [4]
	{ KP6_P5,			0xa5 },	// [5]
	{ KP6_P6,			0xa6 },	// [6]
	{ KP6_P7,			0xa7 },	// [7]
	
	// YB
	{ KP6_P8,			0xb0 },	// [8]
	{ KP6_P9,			0xb1 },	// [9]
	{ KP6_PPERIOD,		0xb2 },	// [.]
	{ KP6_PPLUS,		0xb3 },	// [+]
	{ KP6_PMINUS,		0xb4 },	// [-]
	{ KP6_PMULTIPLY,	0xb5 },	// [*]
	{ KP6_PDIVIDE,		0xb6 },	// [/]
	{ KP6_PRETURN,		0xb7 },	// [RETURN]
};


// P6ޥȥꥯޥå
// (ܸ쥭ܡɤξ)
static const P6MxMap P6MxCode[] = {

	//	  STD   STD+S GRAPH  +S  +S CTRL
	//	--------------------------------------------------
	// Y0
	{ {	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	// CTRL
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	// SHIFT
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	// GRAPH
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
	} },
	
	// Y1
	//	  1     !                   
	{ {	{ 0x31, 0x21, 0x07, 0xe7, 0x00, 0xc7, 0x00, 0x00 },
	//	  q     Q                               Q
		{ 0x71, 0x51, 0x00, 0xe0, 0x00, 0xc0, 0x00, 0x11 },
	//	  a     A                               A
		{ 0x61, 0x41, 0x00, 0xe1, 0x00, 0xc1, 0x00, 0x01 },
	//	  z     Z                           Z
		{ 0x7a, 0x5a, 0x00, 0xe2, 0x8f, 0xc2, 0xaf, 0x1a },
	//	  k     K                               K
		{ 0x6b, 0x4b, 0x00, 0xe9, 0x00, 0xc9, 0x00, 0x0b },
	//	  i     I                             I
		{ 0x69, 0x49, 0x16, 0xe6, 0x00, 0xc6, 0x00, 0x09 },
	//	  8     (     ɴ                
		{ 0x38, 0x28, 0x0d, 0xf5, 0x8d, 0xd5, 0xad, 0x00 },
	//	      <                     
		{ 0x2c, 0x3c, 0x1f, 0xe8, 0xa4, 0xc8, 0xa4, 0x00 }
	} },
	
	// Y2
	//	  2     "                   
	{ {	{ 0x32, 0x22, 0x01, 0xec, 0x00, 0xcc, 0x00, 0x00 },
	//	  w     W                               W
		{ 0x77, 0x57, 0x00, 0xe3, 0x00, 0xc3, 0x00, 0x17 },
	//	  s     S                             S
		{ 0x73, 0x53, 0x0c, 0xe4, 0x00, 0xc4, 0x00, 0x13 },
	//	  x     X                             X
		{ 0x78, 0x58, 0x1c, 0x9b, 0x00, 0xbb, 0x00, 0x18 },
	//	  l     L                             L
		{ 0x6c, 0x4c, 0x1e, 0xf8, 0x00, 0xd8, 0x00, 0x0c },
	//	  o     O                               O
		{ 0x6f, 0x4f, 0x00, 0xf7, 0x00, 0xd7, 0x00, 0x0f },
	//	  9     )                     
		{ 0x39, 0x29, 0x0e, 0xf6, 0x8e, 0xd6, 0xae, 0x00 },
	//	  .     >                     
		{ 0x2e, 0x3e, 0x1d, 0xf9, 0xa1, 0xd9, 0xa1, 0x00 }
	} },
	
	// Y3
	//	  3     #                     
	{ {	{ 0x33, 0x23, 0x02, 0x91, 0x87, 0xb1, 0xa7, 0x00 },
	//	  e     E                         E
		{ 0x65, 0x45, 0x18, 0x92, 0x88, 0xb2, 0xa8, 0x05 },
	//	  d     D                             D
		{ 0x64, 0x44, 0x14, 0x9c, 0x00, 0xbc, 0x00, 0x04 },
	//	  c     C                             C
		{ 0x63, 0x43, 0x1a, 0x9f, 0x00, 0xbf, 0x00, 0x03 },
	//	  ;     +               
		{ 0x3b, 0x2b, 0x82, 0xfa, 0x00, 0xda, 0x00, 0x00 },
	//	  p     P                             P
		{ 0x70, 0x50, 0x10, 0x9e, 0x00, 0xbe, 0x00, 0x10 },
	//	        F6    F1
		{ 0x00, 0xf5, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	  /     ?   ڡ            
		{ 0x2f, 0x3f, 0x80, 0xf2, 0xa5, 0xd2, 0xa5, 0x00 }
	} },
	
	// Y4
	//	  4     $                     
	{ {	{ 0x34, 0x24, 0x03, 0x93, 0x89, 0xb3, 0xa9, 0x00 },
	//	  r     R                             R
		{ 0x72, 0x52, 0x12, 0x9d, 0x00, 0xbd, 0x00, 0x12 },
	//	  f     F                             F
		{ 0x66, 0x46, 0x15, 0xea, 0x00, 0xca, 0x00, 0x06 },
	//	  v     V                             V
		{ 0x76, 0x56, 0x11, 0xeb, 0x00, 0xcb, 0x00, 0x16 },
	//	  :     *    ϡ           
		{ 0x3a, 0x2a, 0x81, 0x99, 0x00, 0xb9, 0x00, 0x00 },
	//	  @                           
		{ 0x40, 0x00, 0x00, 0xde, 0x00, 0xde, 0x00, 0x00 },
	//	        F7    F2
		{ 0x00, 0xf6, 0xf1, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	  _                     
		{ 0x00, 0x5f, 0x83, 0xfb, 0x00, 0xdb, 0x00, 0x00 }
	} },
	
	// Y5
	//	  5     %                     
	{ {	{ 0x35, 0x25, 0x04, 0x94, 0x8a, 0xb4, 0xaa, 0x00 },
	//	  t     T                             T
		{ 0x74, 0x54, 0x19, 0x96, 0x00, 0xb6, 0x00, 0x14 },
	//	  g     G                             G
		{ 0x67, 0x47, 0x13, 0x97, 0x00, 0xb7, 0x00, 0x07 },
	//	  b     B                             B
		{ 0x62, 0x42, 0x1b, 0x9a, 0x00, 0xba, 0x00, 0x02 },
	//	  ]                           
		{ 0x5d, 0x00, 0x85, 0xf1, 0xa3, 0xd1, 0xa3, 0x00 },
	//	  [                            
		{ 0x5b, 0x00, 0x84, 0xdf, 0xa2, 0xdf, 0xa2, 0x00 },
	//	        F8    F3
		{ 0x00, 0xf7, 0xf2, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	  Space Space       Space Space Space Space
		{ 0x20, 0x20, 0x00, 0x20, 0x20, 0x20, 0x20, 0x00 }
	} },
	
	// Y6
	//	  6     &                     
	{ {	{ 0x36, 0x26, 0x05, 0x95, 0x8b, 0xb5, 0xab, 0x00 },
	//	  y     Y     ǯ                        Y
		{ 0x79, 0x59, 0x08, 0xfd, 0x00, 0xdd, 0x00, 0x19 },
	//	  h     H                             H
		{ 0x68, 0x48, 0x0a, 0x98, 0x00, 0xb8, 0x00, 0x08 },
	//	  n     N                               N
		{ 0x6e, 0x4e, 0x00, 0xf0, 0x00, 0xd0, 0x00, 0x0e },
	//	  -     =                   
		{ 0x2d, 0x3d, 0x17, 0xee, 0x00, 0xce, 0x00, 0x00 },
	//	  ^                           
		{ 0x5e, 0x00, 0x00, 0xed, 0x00, 0xcd, 0x00, 0x00 },
	//	        F9    F4
		{ 0x00, 0xf8, 0xf3, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	  0                           
		{ 0x30, 0x00, 0x0f, 0xfc, 0x86, 0xdc, 0xa6, 0x00 }
	} },
	
	// Y7
	//	  7     '                     
	{ {	{ 0x37, 0x27, 0x06, 0xf4, 0x8c, 0xd4, 0xac, 0x00 },
	//	  u     U                               U
		{ 0x75, 0x55, 0x00, 0xe5, 0x00, 0xc5, 0x00, 0x15 },
	//	  j     J                               J
		{ 0x6a, 0x4a, 0x00, 0xef, 0x00, 0xcf, 0x00, 0x0a },
	//	  m     M     ʬ                        M
		{ 0x6d, 0x4d, 0x0b, 0xf3, 0x00, 0xd3, 0x00, 0x0d },
	//
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	                          
		{ 0x5c, 0x00, 0x09, 0xb0, 0x00, 0xb0, 0x00, 0x00 },
	//	        F10   F5
		{ 0x00, 0xf9, 0xf4, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
	} },
	
	// Y8
	//	  RET   RET         RET   RET   RET   RET
	{ {	{ 0x0d, 0x0d, 0x00, 0x0d, 0x0d, 0x0d, 0x0d, 0x00 },
	//	  STOP  STOP  STOP  STOP  STOP  STOP  STOP
		{ 0xfa, 0xfa, 0xfa, 0xfa, 0xfa, 0xfa, 0xfa, 0x00 },
	//	                          
		{ 0x1e, 0x1e, 0x1e, 0x1e, 0x1e, 0x1e, 0x1e, 0x00 },
	//	                          
		{ 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x00 },
	//	                          
		{ 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x00 },
	//	                          
		{ 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x1d, 0x00 },
	//	  TAB   TAB         TAB   TAB   TAB   TAB
		{ 0x09, 0x09, 0x00, 0x09, 0x09, 0x09, 0x09, 0x00 },
	//	  ESC   ESC         ESC   ESC   ESC   ESC
		{ 0x1b, 0x1b, 0x00, 0x1b, 0x1b, 0x1b, 0x1b, 0x00 }
	} },
	
	// Y9
	//	              
	{ {	{ 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	  INS   INS       INS   INS   INS   INS
		{ 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x00 },
	//	  DEL   DEL   ????  DEL   DEL   DEL   DEL
		{ 0x08, 0x08, 0x7f, 0x08, 0x08, 0x08, 0x08, 0x00 },
	//	  PAGE  PAGE  PAGE  PAGE  PAGE  PAGE  PAGE
		{ 0xfe, 0xfe, 0xfc, 0xfe, 0xfe, 0xfe, 0xfe, 0x00 },
	//	  HOME  HOME      HOME  HOME  HOME  HOME
		{ 0x0c, 0x0b, 0x0c, 0x0c, 0x0b, 0x0c, 0x0b, 0x00 },
	//	              MODE
		{ 0x00, 0x00, 0xfd, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//	              CAPS
		{ 0x00, 0x00, 0xfb, 0x00, 0x00, 0x00, 0x00, 0x00 },
	//
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
	} },
	
	// ʹߤȼĥ
	// YA
	//	  [0]   [0]         [0]   [0]   [0]   [0]
	{ {	{ 0x30, 0x30, 0x00, 0x30, 0x30, 0x30, 0x30, 0x00 },
	//	  [1]   [1]         [1]   [1]   [1]   [1]
		{ 0x31, 0x31, 0x00, 0x31, 0x31, 0x31, 0x31, 0x00 },
	//	  [2]   [2]         [2]   [2]   [2]   [2]
		{ 0x32, 0x32, 0x00, 0x32, 0x32, 0x32, 0x32, 0x00 },
	//	  [3]   [3]         [3]   [3]   [3]   [3]
		{ 0x33, 0x33, 0x00, 0x33, 0x33, 0x33, 0x33, 0x00 },
	//	  [4]   [4]         [4]   [4]   [4]   [4]
		{ 0x34, 0x34, 0x00, 0x34, 0x34, 0x34, 0x34, 0x00 },
	//	  [5]   [5]         [5]   [5]   [5]   [5]
		{ 0x35, 0x35, 0x00, 0x35, 0x35, 0x35, 0x35, 0x00 },
	//	  [6]   [6]         [6]   [6]   [6]   [6]
		{ 0x36, 0x36, 0x00, 0x36, 0x36, 0x36, 0x36, 0x00 },
	//	  [7]   [7]         [7]   [7]   [7]   [7]
		{ 0x37, 0x37, 0x00, 0x37, 0x37, 0x37, 0x37, 0x00 }
	} },
	
	// YB
	//	  [8]   [8]         [8]   [8]   [8]   [8]
	{ {	{ 0x38, 0x38, 0x00, 0x38, 0x38, 0x38, 0x38, 0x00 },
	//	  [9]   [9]         [9]   [9]   [9]   [9]
		{ 0x39, 0x39, 0x00, 0x39, 0x39, 0x39, 0x39, 0x00 },
	//	  [.]   [.]         [.]   [.]   [.]   [.]
		{ 0x2e, 0x2e, 0x00, 0x2e, 0x2e, 0x2e, 0x2e, 0x00 },
	//	  [+]   [+]         [+]   [+]   [+]   [+]
		{ 0x2b, 0x2b, 0x00, 0x2b, 0x2b, 0x2b, 0x2b, 0x00 },
	//	  [-]   [-]         [-]   [-]   [-]   [-]
		{ 0x2d, 0x2d, 0x00, 0x2d, 0x2d, 0x2d, 0x2d, 0x00 },
	//	  [*]   [*]         [*]   [*]   [*]   [*]
		{ 0x2a, 0x2a, 0x00, 0x2a, 0x2a, 0x2a, 0x2a, 0x00 },
	//	  [/]   [/]         [/]   [/]   [/]   [/]
		{ 0x2f, 0x2f, 0x00, 0x2f, 0x2f, 0x2f, 0x2f, 0x00 },
	//	  [RET] [RET]       [RET] [RET] [RET] [RET]
		{ 0x0d, 0x0d, 0x00, 0x0d, 0x0d, 0x0d, 0x0d, 0x00 }
	} }
};



enum KeyGroup
{
	KEYSTD = 0,
	KEYSHIFT,
	KEYGRAPH,
	KEYKANA,
	KEYKANASHIFT,
	KEYKKANA,
	KEYKKANASHIFT,
	KEYCTRL
};



////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
KEY6::KEY6( VM6 *vm, const P6ID& id ) : P6DEVENT(vm,id)
{
	ON_SHIFT =			// SHIFT
	ON_GRAPH =			// GRAPH
	ON_KANA  =			// 
	ON_KKANA =			// 
	ON_CTRL  =			// CTRL
	ON_CAPS  = FALSE;	// CAPS
	
	// SDL  -> ۥ Ѵơ֥
	ZeroMemory( &VKTable, sizeof(VKTable) );
	for( int i=0; i < COUNTOF(VKeyDef); i++ ) VKTable[VKeyDef[i].InKey] = VKeyDef[i].VKey;
	
	// ۥ -> P6 Ѵơ֥
	ZeroMemory( &K6Table, sizeof(K6Table) );
//	for( int i=0; i < COUNTOF(P6KeyDef); i++ ) K6Table[P6KeyDef[i].PCKey] = P6KeyDef[i].P6Key;
	
	// P6 -> ޥȥꥯ Ѵơ֥
	ZeroMemory( &MatTable, sizeof(MatTable) );
	for( int i=0; i < COUNTOF(P6KeyMx); i++ ) MatTable[P6KeyMx[i].P6KEY] = P6KeyMx[i].Mat;
	
	// ޥȥꥯ
	memset( P6Matrix,  0xff, NOM*2 );
	memset( P6Mtrx, 0xff, NOM*2 );
	
	// ư̵
	ZeroMemory( &ak, sizeof( AKEY ) );
}

KEY60::KEY60( VM6 *vm, const P6ID& id ) : KEY6(vm,id)
{
	// MODE  CAPS ̵
	MatTable[KP6_MODE] = 0;
	MatTable[KP6_CAPS] = 0;

}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
KEY6::~KEY6( void )
{
	if( ak.Buffer ) delete [] ak.Buffer;
}

KEY60::~KEY60( void ){}


////////////////////////////////////////////////////////////////
// ٥ȥХåؿ
//
// :	id		٥ID
//			clock	å
// :	ʤ
////////////////////////////////////////////////////////////////
void KEY6::EventCallback( int id, int clock )
{
	switch( id ){
	case EID_AKEY:	// ưϽ
		if( !vm->cmt->IsRelay() ){	// TAPEäƤϥߤȯʤ
			BYTE key = GetAutoKey();
			if( key ) vm->subcpu->ReqKeyIntr( 0, key );
		}
		break;
	}
}


////////////////////////////////////////////////////////////////
// ư1ʸ
//   WAIT_AKEY(ܤ10ms)Ȥ˸ƤФ
//
// :	ʤ
// :	BYTE	P6Υ
////////////////////////////////////////////////////////////////
char KEY6::GetAutoKey( void )
{
	// 졼OFFԤ
	if( ak.Relay ){
		if( vm->cmt->IsRelay() ) return 0;
		else                     ak.Relay = FALSE;
	}
	
	// Ԥ?
	if( ak.Wait > 0 ){
		ak.Wait--;
		return 0;
	}
	
	// ʸ
	if( ak.Num-- > 0 ){
		BYTE dat = ak.Buffer[ak.Seek++];
		switch( dat ){
		case 0x17:	// '\w' 
			if( ak.Num-- > 0 ) ak.Wait += (int)((BYTE)ak.Buffer[ak.Seek++]);	// Ԥ
			dat = 0;
			break;
			
		case 0x0a:	// '\r' 졼OFFԤ
			ak.Relay = TRUE;
			dat = 0x0d;
		case 0x0d:	// '\n' ?
			ak.Wait = 15;	// Ԥ15(=150msec)
			break;
			
		default:	// ̤ʸ
			ak.Wait = 1;	// Ԥ1(=10msec)
		}
		return dat;
	}else{
		// Хåեʤ齪λ
		vm->sche->Del( this, EID_AKEY );
		delete [] ak.Buffer;
		ak.Buffer = NULL;
	}
	return 0;
}


////////////////////////////////////////////////////////////////
// ưʸ
//
// :	str		ʸ
//			num		ʸĹ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::SetAutoKey( char *str, int num )
{
	// Хåեö
	if( ak.Buffer ) delete [] ak.Buffer;
	
	// Хåե
	ak.Buffer = new char[num+1];
	if( !ak.Buffer ) return FALSE;
	
	// ʸ򥳥ԡ
	strncpy( ak.Buffer, str, num );
	
	ak.Num    = num;	// Ĥʸ
	ak.Wait   = 30;		// Ԥ(30msec)
	ak.Relay  = FALSE;	// 졼åOFFԤե饰
	ak.Seek   = 0;		// ɹߥݥ
	
	return vm->sche->Add( this, EID_AKEY, WAIT_AKEY, EV_LOOP|EV_MS );
}


////////////////////////////////////////////////////////////////
// ưʸ(ե뤫)
//
// :	filepath	ϥեؤΥѥ
// :	BOOL		TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::SetAutoKeyFile( char *filepath )
{
	FILE *fp;
	char lbuf[1024];
	
	// ưʤ饨顼
	if( ak.Buffer ) return FALSE;
	
	fp = FOPENEN( filepath, "rb" );
	if( !fp ) return FALSE;
	
	fseek( fp, 0, SEEK_END );
	int tsize = ftell( fp ) + 1;
	fseek( fp, 0, SEEK_SET );
	
	// Хåե
	ak.Buffer = new char[tsize+1];
	if( !ak.Buffer ) return FALSE;
	ZeroMemory( ak.Buffer, sizeof(ak.Buffer) );
	
	// ʸɹ
	// ǡ̵ʤޤǷ֤
	while( fgets( lbuf, 1024, fp ) ){
		Sjis2P6( lbuf, lbuf );	// SJIS -> P6
		strcat( ak.Buffer, lbuf );
	}
	tsize = strlen( ak.Buffer );
	
	ak.Num    = tsize;	// Ĥʸ
	ak.Wait   = 30;		// Ԥ(30msec)
	ak.Relay  = FALSE;	// 졼åOFFԤե饰
	ak.Seek   = 0;		// ɹߥݥ
	
	return vm->sche->Add( this, EID_AKEY, WAIT_AKEY, EV_LOOP|EV_MS );
}


////////////////////////////////////////////////////////////////
// 
//
// :	repeat	ԡȤδֳ(ms) 0̵
//			useps2	TRUE:PS2ܡɤȤ FALSE:USBܡɤȤ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::Init( int repeat, BOOL useps2 )
{
	PRINTD( KEY_LOG, "[KEY][Init]\n" );
	
	// ü쥭 ե饰
	ON_SHIFT =
	ON_GRAPH =
	ON_KANA  =
	ON_KKANA =
	ON_CTRL  =
	ON_STOP  =
	ON_CAPS  = FALSE;
	
	// ԡ
	if( repeat ) SDL_EnableKeyRepeat( 500, repeat );
	
	// ܡɼȽ
	KbdType = OSD_GetKeyboardType( useps2 );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 륭ּ
//
// :	ʤ
// :	BYTE	륭ξ 1:ON 0:OFF
//				bit7: ڡ
//				bit6: 0
//				bit5: 
//				bit4: 
//				bit3: 
//				bit2: 
//				bit1: STOP
//				bit0: SHIFT
////////////////////////////////////////////////////////////////
BYTE KEY6::GetKeyJoy( void )
{
	PRINTD( KEY_LOG, "[KEY][GetKeyJoy]\n" );
	
	return	~P6Mtrx[5]&0x80 | ~P6Mtrx[8]&0x20 |  ~P6Mtrx[8]&0x10 | ~P6Mtrx[8]&0x08 |
			~P6Mtrx[8]&0x04 | ~P6Mtrx[8]&0x02 | (~P6Mtrx[0]&0x04)>>2;
}



////////////////////////////////////////////////////////////////
// ޥȥꥯ
//
// :	event	SDLΥϥ٥
// :	ʤ
////////////////////////////////////////////////////////////////
void KEY6::UpdateMatrix( SDL_Event *event )
{
	PRINTD( KEY_LOG, "[KEY][UpdateMatrix]\n" );
	
	// ưϥ̵
	if( ak.Buffer ) return;
	
	switch( event->type ){
	case SDL_KEYDOWN:		// 
	case SDL_KEYUP:
		{
		BYTE KeyPUSH = ( event->type == SDL_KEYDOWN ) ? 1 : 0;	// ե饰
		
		// ̥SDLΥɤ
		int KeyCode = OSD_KeyCodeFix( event->key.keysym.sym, event->key.keysym.scancode, KbdType );
		
		// P6ɤ
		P6KEYsym P6Code = K6Table[VKTable[KeyCode]];
		
		// ޥȥꥯ(бbit0ˤʤ)
		if( MatTable[P6Code] ){
			BYTE matX = 1<<(MatTable[P6Code]&0x0f);
			BYTE matY = (MatTable[P6Code]>>4)&0x0f;
			
			if( KeyPUSH ){
				P6Matrix[matY] &= ~matX;
				
				// ڥԡбλ֡
				// ԡȻϥ꡼󤬽ϤʤΤ
				// Υޥȥꥯ꡼˽񤭴롣
				// ԡȤSDLǤˤƤ֤λ֤
				// CPU¦ǽ褦ˤʤäס
				if( ~P6Matrix[matY+NOM]&matX ) P6Matrix[matY+NOM] |= matX;
			}else{
				P6Matrix[matY] |=  matX;
				//ڥΥμꤳܤɻߡ
				if( P6Matrix[matY+NOM]&matX ) P6Matrix[matY+NOM] &= ~matX;
			}
		}
		
		}
		break;
		
	case SDL_JOYAXISMOTION:
	case SDL_JOYBUTTONDOWN:
	case SDL_JOYBUTTONUP:
		// 祤ƥåξּ
		P6Matrix[NOM-2] = cJoy::GetJoyState();
		break;
	}
}


////////////////////////////////////////////////////////////////
// ޥȥꥯ
//
// :	ʤ
// :	TRUE:Ѳ FALSE:Ѳʤ
////////////////////////////////////////////////////////////////
BOOL KEY6::ScanMatrix( void )
{
	PRINTD( KEY_LOG, "[KEY][ScanMatrix]\n" );
	
	// ޥȥꥯ¸
	memcpy( P6Mtrx, P6Matrix, NOM*2 );
	
	BOOL MatChg  = FALSE;	// ΥޥȥꥯѲե饰
	BOOL KeyPUSH = FALSE;	// ե饰
	BOOL ON_FUNC = FALSE;	// ե󥯥󥭡Ȥե饰
	BYTE MatData = 0;		// P6Υޥȥꥯ(bit7-4:Y-1 bit3-0:X)
	BYTE KeyData = 0;		// P6Υ
	
	// ü쥭Ƚ(ON-OFF֤Ƚ)
	ON_CTRL  = P6Mtrx[0] & 0x02 ? FALSE : TRUE;
	ON_SHIFT = P6Mtrx[0] & 0x04 ? FALSE : TRUE;
	ON_GRAPH = P6Mtrx[0] & 0x08 ? FALSE : TRUE;
	
	// ̥Ƚ
	for( int y=1; (y<(NOM-2))&&~KeyPUSH; y++ ){
		// ΥޥȥꥯѲ?
		if( P6Mtrx[y] != P6Mtrx[y+NOM] ){
			MatChg = TRUE;
			// ?
			if( (P6Mtrx[y] ^ P6Mtrx[y+NOM]) & P6Mtrx[y+NOM] ){
				KeyPUSH = TRUE;
				for( int x=0; x<8; x++ ){
					// ޥȥꥯɥå bit7-4:Y-1 bit3-0:X
					if( (~P6Mtrx[y]>>x)&1 ){
						MatData = y<<4 | x&7;
						break;
					}
				}
			}
		}
	}
	
	if( KeyPUSH ){	// ?
		switch( MatData ){	// ޥȥꥯ
		case 0x96:	// CAPS
			ON_CAPS = ON_CAPS ? FALSE : TRUE;
			ON_FUNC = TRUE;
			break;
			
		case 0x95:	// MODE
			ON_FUNC = TRUE;
			break;
			
		case 0x90:	// 
			ON_KANA = ON_KANA ? FALSE : TRUE;
			ON_FUNC = TRUE;
			break;
			
		case 0x81:	// STOP
			ON_STOP = TRUE;
			break;
			
		case 0x93:	// PAGE or /ڤؤ
			if( ON_SHIFT ){
				ON_KKANA = ON_KKANA ? FALSE : TRUE;
				ON_FUNC  = TRUE;
			}
			break;
			
		case 0x36:			// F01F05(ե󥯥󥭡)
		case 0x46:
		case 0x56:
		case 0x66:
		case 0x76:
			ON_FUNC = TRUE;
			break;
		}
		
		BYTE *KeyMap = (BYTE *)P6MxCode[MatData>>4].Keys[MatData&7];
		if( ON_FUNC ){
			switch( MatData ){	// ޥȥꥯ
			case 0x36:			// F01F05(ե󥯥󥭡)
			case 0x46:
			case 0x56:
			case 0x66:
			case 0x76:
				KeyData = KeyMap[ON_SHIFT ? KEYSHIFT : KEYGRAPH];
				break;
			default:
				KeyData = KeyMap[KEYGRAPH];
			}
		}else{
			if( ON_KANA ){
				if( ON_KKANA ) KeyData = KeyMap[ON_SHIFT ? KEYKKANASHIFT : KEYKKANA];
				else           KeyData = KeyMap[ON_SHIFT ? KEYKANASHIFT  : KEYKANA];
			}else{
				if( ON_GRAPH )
					KeyData = KeyMap[KEYGRAPH];
				else
					if( ON_CTRL ) KeyData = KeyMap[KEYCTRL];
					else          KeyData = KeyMap[ON_SHIFT ? KEYSHIFT : KEYSTD];
			}
		}
		
		// ߽
		if( KeyData ){	// ͭʥ?
			// flag
			// bit 0 : ON_STOP
			// bit 1 : ON_GRAPH
			// bit 2 : ON_FUNC
			
			vm->subcpu->ReqKeyIntr( ON_STOP ? 1 : 0 | ON_GRAPH ? 2 : 0 | ON_FUNC ? 4 : 0, KeyData );
			ON_STOP = FALSE;
		}
	}
	
	// 祤ƥåȽ
	for( int y=NOM-2; y<NOM; y++ ){
		// ΥޥȥꥯѲ?
		if( P6Mtrx[y] != P6Mtrx[y+NOM] ) MatChg = TRUE;
	}
	
	// ѲХޥȥꥯ¸
	if( MatChg ) memcpy( P6Matrix+NOM, P6Mtrx, NOM );
	
	return MatChg;
}


////////////////////////////////////////////////////////////////
// ޥȥꥯ
//
// :	ʤ
// :	int		ޥȥꥯ
////////////////////////////////////////////////////////////////
int KEY6::GetMatrixSize( void )
{
	return NOM*2;
}


////////////////////////////////////////////////////////////////
// ޥȥꥯݥ󥿼
//
// :	ʤ
// :	BYTE *	ޥȥꥯǡؤΥݥ
////////////////////////////////////////////////////////////////
BYTE *KEY6::GetMatrix( void )
{
	return P6Matrix;
}


////////////////////////////////////////////////////////////////
// ޥȥꥯݥ(¸)
//
// :	ʤ
// :	BYTE *	ޥȥꥯǡ(¸)ؤΥݥ
////////////////////////////////////////////////////////////////
BYTE *KEY6::GetMatrix2( void )
{
	return P6Mtrx;
}


////////////////////////////////////////////////////////////////
// ܡɥ󥸥ּ
//
// :	ʤ
// :	BYTE	󥸥
//				KI_KANA:  ʥ⡼ON
//				KI_KKANA: ʥ⡼ON
//				KI_CAPS:  CAPS ON
////////////////////////////////////////////////////////////////
BYTE KEY6::GetKeyIndicator( void )
{
	PRINTD( KEY_LOG, "[KEY][GetKeyIndicator] -> " );
	
	BYTE ret = 0;
	
	// ʥ
	if( ON_KANA ) ret |= ON_KKANA ? KI_KKANA : KI_KANA;
	
	// CAPS
	if( ON_CAPS ) ret |= KI_CAPS;
	
	PRINTD1( KEY_LOG, "%d\n", ret );
	
	return ret;
}


////////////////////////////////////////////////////////////////
// ۥ -> P6 
//
// :	VKeyConv *	ۥ -> P6¤ Ƭݥ
// :	ʤ
////////////////////////////////////////////////////////////////
void KEY6::SetVKeySymbols( VKeyConv *vcnv )
{
	while( vcnv->PCKey != KVC_LAST ){
		K6Table[vcnv->PCKey] = vcnv->P6Key;
		vcnv++;
	}
}


////////////////////////////////////////////////////////////////
// 祤ƥåּ
//
// :	ʤ
// :	BYTE	祤ƥåξ 1:OFF 0:ON <-!
//				bit4: ܥ
//				bit3: 
//				bit2: 
//				bit1: 
//				bit0: 
////////////////////////////////////////////////////////////////
BYTE KEY6::GetJoy( void )
{
	return P6Mtrx[NOM-2];
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
//
// :	Ini		INI֥ȥݥ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::DokoSave( cIni *Ini )
{
	cSche::evinfo e;
	char stren[16];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "KEY", NULL, "ON_KANA",		"%s",	ON_KANA  ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_KKANA",		"%s",	ON_KKANA ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_STOP",		"%s",	ON_STOP  ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_CAPS",		"%s",	ON_CAPS  ? "Yes" : "No" );
	
	if( ak.Buffer && ak.Num > 0 ){
		char strva[strlen(ak.Buffer)*2];
		for( int i=0; i<ak.Num; i++ ) sprintf( &strva[i*2], "%02X", ak.Buffer[i] );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Buffer",		"%s",	strva );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Wait",		"%d",	ak.Wait );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Relay",		"%s",	ak.Relay ? "Yes" : "No" );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Seek",		"%d",	ak.Seek );
	}
	
	// ٥
	e.id = EID_AKEY;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "KEY", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɤǤLOAD
//
// :	Ini		INI֥ȥݥ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::DokoLoad( cIni *Ini )
{
	int yn;
	cSche::evinfo e;
	char stren[16],strva[256];
	char strrs[64];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->GetTruth( "KEY", "ON_KANA",	&ON_KANA,	ON_KANA );
	Ini->GetTruth( "KEY", "ON_KKANA",	&ON_KKANA,	ON_KKANA );
	Ini->GetTruth( "KEY", "ON_STOP",	&ON_STOP,	ON_STOP );
	Ini->GetTruth( "KEY", "ON_CAPS",	&ON_CAPS,	ON_CAPS );
	
	if( Ini->GetString( "TAPE", "AKEY_Buffer", strva, "" ) ){
		int num = strlen( strva ) / 2;
		char strak[num+1];
		for( int i=0; i<num; i++ ){
			char dt[5] = "0x";
			strncpy( &dt[2], &strva[i*2], 2 );
			strak[i] = strtol( dt, NULL, 16 );
		}
		SetAutoKey( strak, num );
		Ini->GetInt(   "TAPE",	"AKEY_Wait",		&ak.Wait,		ak.Wait );
		Ini->GetTruth( "TAPE",	"AKEY_Relay",		&ak.Relay,		ak.Relay );
		Ini->GetInt(   "TAPE",	"AKEY_Seek",		&ak.Seek,		ak.Seek );
	}
	
	// ٥
	e.id = EID_AKEY;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "KEY", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	return TRUE;
}
