#include <stdlib.h>
#include <new>

#include <SDL_endian.h>

#include "mc6847.h"
#include "common.h"  // add for Sjis2UTF8 Windy 2007/4/21
#include "error.h"
#include "memory.h"
#include "osd.h"
#include "schedule.h"
#include "z80.h"


// ٥ID
#define	EID_VSYNC	(1)		/* ľƱ */
#define	EID_HSYNC	(2)		/* ʿƱ */

#define	VSLINE		(262)	/* ľȡ饤 */
#define	HSDCLOCK	(455)	/* ʿȡ(ɥåȥå) */
#define	VLINES60	(192)	/* ľɽ饤(N60)  */
#define	VLINES62	(200)	/* ľɽ饤(N60m) */
#define	HCLOCK60	(296)	/* ʿɽ(N60)  */
#define	HCLOCK62	(368)	/* ʿɽ(N60m) */

#define	P6WIDTH		(256)
#define	P6HEIGHT	(192)
#define	P2WIDTH		(320)
#define	P2HEIGHT	(200)

// ܡ
#define	LB60		((int)((WBBUF-P6WIDTH)/2))
#define	RB60		(WBBUF-P6WIDTH-LB60)
#define	TB60		((int)((HBBUF-P6HEIGHT)/2))
#define	BB60		(HBBUF-P6HEIGHT-TB60)
#define	LB62		((int)((WBBUF-P2WIDTH)/2))
#define	RB62		(WBBUF-P2WIDTH-LB62)
#define	TB62		((int)((HBBUF-P2HEIGHT)/2))
#define	BB62		(HBBUF-P2HEIGHT-TB62)


// 顼å
static BYTE COL_MODE1[5];		// mode 1 -----
static BYTE COL_MODE2[9];		// mode 2 -----
static BYTE COL_MODE3[10][4];	// mode 3 -----
static BYTE COL_MODE4[2][2];	// mode 4 -----


// --------- 60 ---------
static const BYTE COL60_MODE1[] =		// mode 1 -----
				{ 17,18,19,20,16 };

static const BYTE COL60_MODE2[] =		// mode 2 -----
				{ 21,22,23,24,25,26,27,28,16 };

static const BYTE COL60_MODE3[][4] = {	// mode 3 -----
				{ 29,30,31,32 },
				{ 33,34,35,36 },
				{ 37,41,42,38 },	// mode4
				{ 39,45,46,40 },	// mode4
				{ 37,42,41,38 },	// mode4
				{ 39,46,45,40 },	// mode4
				{ 37,43,44,38 },	// mode4
				{ 39,47,48,40 },	// mode4
				{ 37,44,43,38 },	// mode4
				{ 39,48,47,40 }		// mode4
			};

static const BYTE COL60_MODE4[][2] = {	// mode 4 -----
				{ 37,38 },
				{ 39,40 }
			};

//static const BYTE COL60_MODE4A[][14] = {	// mode 4 ˤ -----
//				{ 37,38,37,38,37,38,37,38,37,38,37,38,37,38 },
//				{ 39,40,65,66,67,68,69,70,71,72,73,74,75,76 }};
//			};

// --------- mk2 ---------
static const BYTE COL62_MODE1[] =		// mode 1-1 -----
				{ 64,57,59,57,57 };

static const BYTE COL62_MODE2[] =		// mode 1-2 -----
				{ 59,60,61,58,64,63,62,50,57 };

static const BYTE COL62_MODE3[][4] = {	// mode 1-3 -----
				{ 59,60,61,58 },
				{ 64,63,62,50 },
				{ 57,41,42,59 },	// mode1-4
				{ 57,45,46,64 },	// mode1-4
				{ 57,42,41,59 },	// mode1-4
				{ 57,46,45,64 },	// mode1-4
				{ 57,43,44,59 },	// mode1-4
				{ 57,47,48,64 },	// mode1-4
				{ 57,44,43,59 },	// mode1-4
				{ 57,48,47,64 }		// mode1-4
			};

static const BYTE COL62_MODE4[][2] = {	// mode 1-4 -----
				{ 57,59 },
				{ 57,64 }
			};

static const BYTE COL_MOD21[] =			// mode 2-1,2 -----
				{ 49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64 };

static const BYTE COL_MOD23[][16] = {	// mode 2-3 -----
				{ 49,53,50,54,51,55,52,56,57,61,58,62,59,63,60,64 },
				{ 59,60,61,58,64,63,62,50,59,60,61,58,64,63,62,50 }
			};

static const BYTE COL_MOD24[][16] = {	// mode 2-4 -----
				{ 49,53,50,54,51,55,52,56,57,61,58,62,59,63,60,64 },
				{ 59,60,61,58,64,63,62,50,59,60,61,58,64,63,62,50 }
			};



////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
VDG6::VDG6( VM6 *vm, const P6ID& id ) : P6DEVENT(vm,id)
{
	bbuf     = NULL;
	
	Mode4Col = DEFAULT_MODE4_COLOR;
	CrtDisp  = TRUE;
	N60Win   = TRUE;
	
	VramAddr = 0xc200;
	AttrAddr = 0xc000;
	
	BCol     = 0;
	
	VLines   = VLINES60;
	VLcnt    = VLINES60;
	HWait    = 0;
	
	// ѥåȽ
	for( int i=0; i<128; i++ ){
		// ɸ५顼
		GPal[i].r = vm->cGlobal::UserColor[i].r;
		GPal[i].g = vm->cGlobal::UserColor[i].g;
		GPal[i].b = vm->cGlobal::UserColor[i].b;
		// 饤ѥ顼
		GPal[i+128].r = ( GPal[i].r * vm->cGlobal::ScanLineBr ) / 100;
		GPal[i+128].g = ( GPal[i].g * vm->cGlobal::ScanLineBr ) / 100;
		GPal[i+128].b = ( GPal[i].b * vm->cGlobal::ScanLineBr ) / 100;
	}
}

VDG60::VDG60( VM6 *vm, const ID& id ) : VDG6(vm,id), Device(id)
{
	CGRomV = NULL;
}

VDG62::VDG62( VM6 *vm, const ID& id ) : VDG6(vm,id), Device(id)
{
	Mk2CharMode  = TRUE;
	Mk2GraphMode = FALSE;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
VDG6::~VDG6( void )
{
	if( bbuf ) SDL_FreeSurface( bbuf );
}

VDG60::~VDG60( void )
{
	if( CGRomV ) delete [] CGRomV;
}

VDG62::~VDG62( void ){}


////////////////////////////////////////////////////////////////
// ٥ȥХåؿ
//
// :	id		٥ID
//			clock	å
// :	ʤ
////////////////////////////////////////////////////////////////
void VDG6::EventCallback( int id, int clock )
{
	switch( id ){
	case EID_VSYNC:
		// VSYNCΤ
		vm->sche->OnVSYNC();
		VLcnt = VLines;		// ɽ饤󥫥Ƚ
		vm->sche->Reset( this, EID_HSYNC );
		break;
		
	case EID_HSYNC:
		// ɽ֤ǤBUSREQ˱ƥȤ򤫤
		if( VLcnt ){
			vm->cpu->AddWait( HWait );
			
			VLcnt--;
		}
		break;
	}
}


////////////////////////////////////////////////////////////////
// ȥӥ塼ȥǡ
////////////////////////////////////////////////////////////////
BYTE VDG6::GetAttr( WORD addr )
{
	return vm->mem->ReadMainRam( AttrAddr + addr );
}


////////////////////////////////////////////////////////////////
// VRAMǡ
////////////////////////////////////////////////////////////////
BYTE VDG6::GetVram( WORD addr )
{
	return vm->mem->ReadMainRam( VramAddr + addr );
}


////////////////////////////////////////////////////////////////
// Font1ǡ
////////////////////////////////////////////////////////////////
BYTE VDG6::GetFont1( WORD addr )
{
	return vm->mem->ReadCGrom1( addr );
}


////////////////////////////////////////////////////////////////
// Font2ǡ
////////////////////////////////////////////////////////////////
BYTE VDG62::GetFont2( WORD addr )
{
	return vm->mem->ReadCGrom2( addr );
}


////////////////////////////////////////////////////////////////
// VDG Fontǡ
////////////////////////////////////////////////////////////////
BYTE VDG60::GetFontV( WORD addr )
{
	if( CGRomV ) return CGRomV[addr&0x3ff];
	else         return 0xff;
}


////////////////////////////////////////////////////////////////
// CRTɽ
////////////////////////////////////////////////////////////////
void VDG6::SetCrtDisp( BOOL st )
{
	CrtDisp = st;
	
	// BUSREQѲ
	ChangeBUSREQ();
}


////////////////////////////////////////////////////////////////
// BUSREQѲ
////////////////////////////////////////////////////////////////
void VDG6::ChangeBUSREQ( void )
{
	VLines = N60Win ? VLINES60 : VLINES62;
	
	// ʿɽ֥Ȥ׻
	if( CrtDisp ) HWait = (int)((double)(vm->sche->GetClock() / VSYNC_HZ / VSLINE * ( N60Win ? HCLOCK60 : HCLOCK62 ) / HSDCLOCK));
	else          HWait = 0;
}


////////////////////////////////////////////////////////////////
// ե˥ѥå
////////////////////////////////////////////////////////////////
BOOL VDG6::SetPalette( SDL_Surface *scr )
{
	PRINTD( GRP_LOG, "[VDG][SetPalette]\n" );
	
	if( SDL_SetColors( scr, GPal, 0, 255 ) ) return TRUE;
	else                                     return FALSE;
}


////////////////////////////////////////////////////////////////
// ХåХåե
////////////////////////////////////////////////////////////////
BOOL VDG6::UpdateBackBuf( void )
{
	PRINTD( GRP_LOG, "[VDG][UpdateBackBuf]\n" );
	
	if( !RefreshBackBuf() ) return FALSE;
	Vram2Screen();
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ХåХåեեå
////////////////////////////////////////////////////////////////
BOOL VDG6::RefreshBackBuf( void )
{
	PRINTD( GRP_LOG, "[VDG][RefreshBackBuf]\n" );
	
	// Ѥ?
	if( bbuf ){
		// ѤäƤʤOK
		// ѤäƤϰö
		if( (bbuf->w == WBBUF) && (bbuf->h == HBBUF) ) return TRUE;
		else                                           SDL_FreeSurface( bbuf );
	}
	
	// ХåХåե
	bbuf = SDL_CreateRGBSurface( SDLOP_SURFACE, WBBUF, HBBUF, 8, 0, 0, 0, 0 );
	if( !bbuf ) return FALSE;
	
	// ѥå
	if( !SetPalette( bbuf ) ) return FALSE;
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL VDG6::Init( void )
{
	// ٥ɲ
	if( !vm->sche->Add( this, EID_VSYNC, VSYNC_HZ,        EV_LOOP|EV_HZ ) ) return FALSE;
	if( !vm->sche->Add( this, EID_HSYNC, VSYNC_HZ*VSLINE, EV_LOOP|EV_HZ ) ) return FALSE;
	
	// ХåХåեեå
	return RefreshBackBuf();
}

BOOL VDG60::Init( void )
{
	PRINTD( GRP_LOG, "[VDG][60][Init]\n" );
	
	// 顼
	for( int i=0; i<5; i++ ) COL_MODE1[i] = COL60_MODE1[i];
	for( int i=0; i<9; i++ ) COL_MODE2[i] = COL60_MODE2[i];
	for( int i=0; i<10; i++ )
		for( int j=0; j<4; j++ )
			COL_MODE3[i][j] = COL60_MODE3[i][j];
	for( int i=0; i<2; i++ )
		for( int j=0; j<2; j++ )
			COL_MODE4[i][j] = COL60_MODE4[i][j];
	
	// 
	return VDG6::Init();
}

BOOL VDG62::Init( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Init]\n" );
	
	// 顼
	for( int i=0; i<5; i++ ) COL_MODE1[i] = COL62_MODE1[i];
	for( int i=0; i<9; i++ ) COL_MODE2[i] = COL62_MODE2[i];
	for( int i=0; i<10; i++ )
		for( int j=0; j<4; j++ )
			COL_MODE3[i][j] = COL62_MODE3[i][j];
	for( int i=0; i<2; i++ )
		for( int j=0; j<2; j++ )
			COL_MODE4[i][j] = COL62_MODE4[i][j];
	
	// 
	return VDG6::Init();
}


////////////////////////////////////////////////////////////////
// VDGեɹ
////////////////////////////////////////////////////////////////
BOOL VDG60::LoadFont( char *path, char *fname )
{
	PRINTD2( GRP_LOG, "[VDG][60][LoadFont] %s %s\n", path, fname );
	
	// ö
	if( CGRomV ) delete [] CGRomV;
	CGRomV = NULL;
	
	try{
		// 
		CGRomV = new BYTE[0x400];	// 1KB
		
		// ե뤫ɹ
		char fpath[PATH_MAX] = "";
		if( path ){
			strncpy( fpath, path, PATH_MAX );
			AddDelimiter( fpath );
		}
		strncat( fpath, fname, PATH_MAX );
		
		FILE *fp = FOPENEN( fpath, "rb" );
		if( !fp ) throw Error::NoRom;
		DWORD LoadSize = fread( CGRomV, sizeof(BYTE), 1024, fp );
		fclose( fp );
		
		// å
		if( LoadSize != 1024 ) throw Error::RomSizeNG;
	}
	// new ˼Ԥ
	catch( std::bad_alloc ){
//		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	// 㳰ȯ
	catch( Error::Errno i ){
//		Error::SetError( i );
		
		switch(i){
		case Error::NoRom:		// եΥץ˼Ԥ
		case Error::RomSizeNG:	// ʤ
		default:				// 
			delete [] CGRomV;
			CGRomV = NULL;
		}
		return FALSE;
	}
	
	return TRUE;
}




////////////////////////////////////////////////////////////////
// CRT KILL
////////////////////////////////////////////////////////////////
void VDG60::CrtKill( void )
{
	PRINTD( GRP_LOG, "[VDG][CrtKill]\n" );
	
	SDL_Rect darea = { LB60, TB60, P6WIDTH, P6HEIGHT };
	BYTE adat = GetAttr(0);
	BYTE bcol;
	BYTE css = (adat>>1)&1;
	if( adat&0x80 ){	// եå
		if( adat&0x10 ){	// Υ
			SDL_FillRect( bbuf, NULL, COL_MODE4[css][1] );
			bcol = COL_MODE4[css][0];
			if( adat&0x0c == 0x0c ){	// 256X192
				for( int y=darea.y; y<darea.y+darea.h; y++ ){
					BYTE *sur = (BYTE *)bbuf->pixels + bbuf->pitch * y + darea.x;
					for( int x=0; x<darea.w; x+=8 ){
						BYTE d = rand();
						d = (d>>4)^d;
						for( int i=0; i<8; i++ ){
							*sur++ = COL_MODE4[css][d&1];
							d >>= 1;
						}
					}
				}
			}else{						// 128X64,96,128
				for( int y=darea.y; y<darea.y+darea.h; y++ ){
					BYTE *sur = (BYTE *)bbuf->pixels + bbuf->pitch * y + darea.x;
					for( int x=0; x<darea.w; x+=8 ){
						BYTE d = rand();
						d = (d>>4)^d;
						for( int i=0; i<8; i+=2 ){
							*sur++ = COL_MODE4[css][d&1];
							*sur++ = COL_MODE4[css][d&1];
							d >>= 1;
						}
					}
				}
			}
		}else{				// 顼
			SDL_FillRect( bbuf, NULL, COL_MODE3[css][0] );
			for( int y=darea.y; y<darea.y+darea.h; y++ ){
				BYTE *sur = (BYTE *)bbuf->pixels + bbuf->pitch * y + darea.x;
				for( int x=0; x<darea.w; x+=8 ){
					BYTE d = rand();
					d = (d>>4)^d;
					for( int i=0; i<8; i+=2 ){
						*sur++ = COL_MODE3[css][d&3];
						*sur++ = COL_MODE3[css][d&3];
						d >>= 2;
					}
				}
			}
		}
	}else{				// ե˥塼å,ߥեå
		SDL_FillRect( bbuf, NULL,   COL_MODE1[4] );
		SDL_FillRect( bbuf, &darea, COL_MODE1[css ? 3 : 1] );
	}
}

void VDG62::CrtKill( void )
{
	PRINTD( GRP_LOG, "[VDG][CrtKill]\n" );
	
	// mk2ʹߤξ
	//  RGBϤξϿù?
	//  ӥǥϤξϺ
	SDL_Rect darea = { LB62, TB62, P2WIDTH, P2HEIGHT };
	SDL_FillRect( bbuf, NULL,   COL_MODE1[4] );
	SDL_FillRect( bbuf, &darea, COL_MODE1[1] );
	
	// ξ
//	for( int y=TB62; y<TB62+P2HEIGHT; y++ ){
//		BYTE *sur = (BYTE *)bbuf->pixels + bbuf->pitch * y + LB62;
//		for( int x=0; x<P2WIDTH; x+=8 ){
//			BYTE d = rand();
//			for( int i=0; i<8; i++ ){
//				*sur++ = d&1 ? COL_MODE1[0] : COL_MODE1[1];
//				d >>= 1;
//			}
//		}
//	}
}


////////////////////////////////////////////////////////////////
// ⡼ѥ֥롼
////////////////////////////////////////////////////////////////

// ----------------------------------
//     PC-6001
// ----------------------------------

// ⡼1 1ʸPUT
void VDG6::Putc_M1( int x, int y, WORD addr )
{
	PRINTD3( GRP_LOG, "[VDG][Putc_M1] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr,fdat;
	DWORD fg,bg;
	DWORD *doff;
	WORD soff;
	
	data = GetVram( addr );
	attr = GetAttr( addr );
	
	doff = (DWORD *)bbuf->pixels + LB60/sizeof(DWORD) + TB60*bbuf->pitch/sizeof(DWORD) + x*2 + y*bbuf->pitch*12/sizeof(DWORD);
	soff = data*16;
	fg   = COL_MODE1[attr&3];
	bg   = COL_MODE1[attr&2 | ~attr&1];
	for( int yc=0; yc<12; yc++ ){
		fdat    = GetFont1( soff++ );
		*doff++ = BTODW( fdat&0x80 ? fg : bg, fdat&0x40 ? fg : bg, fdat&0x20 ? fg : bg, fdat&0x10 ? fg : bg );
		*doff   = BTODW( fdat&0x08 ? fg : bg, fdat&0x04 ? fg : bg, fdat&0x02 ? fg : bg, fdat&0x01 ? fg : bg );
		
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// ⡼2-1 1ʸPUT
void VDG6::Putc_M2( int x, int y, WORD addr)
{
	PRINTD3( GRP_LOG, "[VDG][Putc_M2] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr;
	DWORD fg,bg,d1,d2;
	DWORD *doff;
	
	data = GetVram( addr );
	attr = GetAttr( addr );
	
	doff = (DWORD *)bbuf->pixels + LB60/sizeof(DWORD) + TB60*bbuf->pitch/sizeof(DWORD) + x*2 + y*bbuf->pitch*12/sizeof(DWORD);
	fg   = COL_MODE2[(attr<<1)&4 | (data>>6)&3];
	fg  |= fg<<8 | fg<<16 | fg<<24 ;
	bg   = COL_MODE2[8];
	bg  |= bg<<8 | bg<<16 | bg<<24 ;
	
	d1 = data&0x20? fg: bg;
	d2 = data&0x10? fg: bg;
	for( int yc=0; yc<4; yc++ ){
		*doff++ = d1;
		*doff   = d2;
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
	
	d1 = data&0x08? fg: bg;
	d2 = data&0x04? fg: bg;
	for( int yc=4; yc<8; yc++ ){
		*doff++ = d1;
		*doff   = d2;
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
	
	d1 = data&0x02? fg: bg;
	d2 = data&0x01? fg: bg;
	for( int yc=8; yc<12; yc++ ){
		*doff++ = d1;
		*doff   = d2;
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// ⡼0 1ʸPUT
void VDG60::Putc_M0( int x, int y, WORD addr )
{
	PRINTD3( GRP_LOG, "[VDG][Putc_M0] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr,fdat;
	DWORD fg,bg;
	DWORD *doff;
	WORD soff;
	
	data = GetVram( addr ) & 0x3f;
	attr = GetAttr( addr );
	
	doff = (DWORD *)bbuf->pixels + LB60/sizeof(DWORD) + TB60*bbuf->pitch/sizeof(DWORD) + x*2 + y*bbuf->pitch*12/sizeof(DWORD);
	soff = data*16;
	fg   = COL_MODE1[attr&3];
	bg   = COL_MODE1[attr&2 | ~attr&1];
	for( int yc=0; yc<12; yc++ ){
		fdat    = GetFontV( soff++ );
		*doff++ = BTODW( fdat&0x80 ? fg : bg, fdat&0x40 ? fg : bg, fdat&0x20 ? fg : bg, fdat&0x10 ? fg : bg );
		*doff   = BTODW( fdat&0x08 ? fg : bg, fdat&0x04 ? fg : bg, fdat&0x02 ? fg : bg, fdat&0x01 ? fg : bg );
		
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// ⡼2-2 1ʸPUT
void VDG60::Putc_M21( int x, int y, WORD addr)
{
	PRINTD3( GRP_LOG, "[VDG][Putc_M21] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr;
	DWORD fg,bg,d1,d2;
	DWORD *doff;
	
	data = GetVram( addr );
	attr = GetAttr( addr );
	
	doff = (DWORD *)bbuf->pixels + LB60/sizeof(DWORD) + TB60*bbuf->pitch/sizeof(DWORD) + x*2 + y*bbuf->pitch*12/sizeof(DWORD);
	fg   = COL_MODE2[(data>>4)&7];
	fg  |= fg<<8 | fg<<16 | fg<<24;
	bg   = COL_MODE2[8];
	bg  |= bg<<8 | bg<<16 | bg<<24 ;
	
	d1 = data&0x08? fg: bg;
	d2 = data&0x04? fg: bg;
	for( int yc=0; yc<6; yc++ ){
		*doff++ = d1;
		*doff   = d2;
		doff += bbuf->pitch/sizeof(DWORD)-1;
	}
	
	d1 = data&0x02? fg: bg;
	d2 = data&0x01? fg: bg;
	for( int yc=6; yc<12; yc++ ){
		*doff++ = d1;
		*doff   = d2;
		doff += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// ----------------------------------
//     Mode 1,2 sub
// ----------------------------------
void VDG60::Mode_12( void )
{
	PRINTD( GRP_LOG, "[VDG][60][Mode_12]\n" );
	
	WORD AVOFF = 0;
	
	// ܡ顼ɤĤ֤
	SDL_FillRect( bbuf, NULL, COL_MODE1[4] );
	
	for( int y=0; y<16; y++ ){
		for( int x=0; x<32; x++ ){
			switch( GetAttr(AVOFF)&0x60 ){
			case 0x00:			// *** ⡼0 ***
				if( CGRomV ){
					// եȤ
					Putc_M0( x, y, AVOFF );
					break;
				}
			case 0x20:			// *** ⡼1 ***
				Putc_M1( x, y, AVOFF );
				break;
			case 0x40:			// *** ⡼2-2 ***
				Putc_M21( x, y, AVOFF );
				break;
			case 0x60:			// *** ⡼2-1 ***
				Putc_M2( x, y, AVOFF );
				break;
			}
			AVOFF++;
		}
	}
}


// ----------------------------------
//     Mode 3,4 sub
// ----------------------------------
void VDG60::Mode_34( void )
{
	PRINTD( GRP_LOG, "[VDG][60][Mode_34]\n" );
	
	// 嵡VDG(ղϩ?)ˤϥХ餷
	// ȥӥ塼ȥɥ쥹Ф顼åȤɤ뤬
	// ⡼Ƚλˤ1byteΥɥ쥹ɤǤޤ褦
	// mk2,66ǤϤɤɤ
	
	int AOff  = 0;	// ȥӥ塼ȥեå;
	int VOff  = 0;	// VRAMեå
	int VOffL = 0;	// VRAMեå ¸
	BYTE Vram;		// VRAMǡ
	int Css;		// 顼å
	int Gmode;		// եå⡼(07)
	int GmodeL;		// եå⡼(07) ¸
	int ACnt;		// ȥӥ塼ȥ
	int HLine = 0;	// 饤
	
	DWORD *doff = (DWORD *)bbuf->pixels;
	
	// ¦ܡ
	for( int i=0; i<(int)(bbuf->pitch/sizeof(DWORD)*TB60); i++ ) *(doff++) = BCol;
	
	for( int y=0; y<P6HEIGHT; y++ ){
		Gmode = GmodeL = (GetAttr(AOff+1)>>2)&7;	// եå⡼(07)
		ACnt  = 0;		// ȥӥ塼ȥ󥿽
		int DWidth  = P6WIDTH/sizeof(DWORD);
		
		// 饤Ƚ
		if( !HLine ){
			VOffL = VOff;	// VRAMեå¸
			
			switch( Gmode ){
			case 0:		//  64 64 顼
			case 2:		// 128 64 顼
			case 4:		// 128 64 Υ
				HLine = 3;
				break;
			case 1:		// 128 96 顼
			case 6:		// 128 96 Υ
				HLine = 2;
				break;
			case 3:		// 128192 顼
			case 5:		// 128192 Υ
			case 7:		// 256192 Υ
				HLine = 1;
			}
		}
		
		// ¦ܡ
		for( int i=0; i<(int)(LB60/sizeof(DWORD)); i++ ) *(doff++) = BCol;
		
		while( DWidth > 0 ){
			DWORD d;
			BYTE b1,b2,b3,b4;
			
			Css     = (GetAttr(AOff+ACnt)>>1)&1;	// 顼å
			++ACnt &= 0x1f;	// ȥӥ塼ȥ󥿤32bytesǥ롼
			Gmode   = (GetAttr(AOff+ACnt)>>2)&7;	// եå⡼(07)
			Vram    = GetVram( VOff++ );			// VRAMǡ
			
			// ()ե˥塼åߤƤν
			if( !(GetAttr(AOff+ACnt)&0x80) ) Gmode = -1;
			
			switch( Gmode ){
			case -1:	//  ()ե˥塼å
				b1 = COL_MODE1[Css<<1];
				d  = BTODW( b1, b1, b1, b1 );
				
				*(doff++) = d;
				*(doff++) = d;
				DWidth -= 2;
				break;
				
			case 0:		//  64 64 顼
				for( int i=3; i>=0; i-- ){
					b1 = COL_MODE3[Css][(Vram>>(i*2))&3];
					*(doff++) = BTODW( b1, b1, b1, b1 );
				}
				DWidth -= 4;
				break;
				
			case 1:		// 128 96 顼
			case 2:		// 128 64 顼
			case 3:		// 128192 顼
				b1 = COL_MODE3[Css][(Vram>>6)&3];
				b2 = COL_MODE3[Css][(Vram>>4)&3];
				*(doff++) = BTODW( b1, b1, b2, b2 );
				b3 = COL_MODE3[Css][(Vram>>2)&3];
				b4 = COL_MODE3[Css][ Vram    &3];
				*(doff++) = BTODW( b3, b3, b4, b4 );
				
				DWidth -= 2;
				break;
				
			case 4:		// 128 64 Υ
			case 5:		// 128192 Υ
			case 6:		// 128 96 Υ
				for( int i=3; i>=0; i-- ){
					b1 = COL_MODE4[Css][(Vram>>(i*2+1))&1];
					b2 = COL_MODE4[Css][(Vram>>(i*2  ))&1];
					*(doff++) = BTODW( b1, b1, b2, b2 );
				}
				DWidth -= 4;
				break;
				
			case 7:		// 256192 Υ
				if( Mode4Col ){		// 顼ɽ
					// 128192 顼
					int CsC = Css + Mode4Col*2;
					b1 = COL_MODE3[CsC][(Vram>>6)&3];
					b2 = COL_MODE3[CsC][(Vram>>4)&3];
					*(doff++) = BTODW( b1, b1, b2, b2 );
					b3 = COL_MODE3[CsC][(Vram>>2)&3];
					b4 = COL_MODE3[CsC][ Vram    &3];
					*(doff++) = BTODW( b3, b3, b4, b4 );
				}else{
					// 256192 Υ
					b1 = COL_MODE4[Css][(Vram>>7)&1];
					b2 = COL_MODE4[Css][(Vram>>6)&1];
					b3 = COL_MODE4[Css][(Vram>>5)&1];
					b4 = COL_MODE4[Css][(Vram>>4)&1];
					*(doff++) = BTODW( b1, b2, b3, b4 );
					b1 = COL_MODE4[Css][(Vram>>3)&1];
					b2 = COL_MODE4[Css][(Vram>>2)&1];
					b3 = COL_MODE4[Css][(Vram>>1)&1];
					b4 = COL_MODE4[Css][ Vram    &1];
					*(doff++) = BTODW( b1, b2, b3, b4 );
				}
				DWidth -= 2;
				break;
			}
		}
		
		// ܡ顼Ƚ
		if     ( Gmode < 0 ) BCol = COL_MODE1[4];
		else if( Gmode < 4 ) BCol = COL_MODE3[Css][0];
		else                 BCol = COL_MODE4[Css][1];
		BCol |= BCol<< 8;
		BCol |= BCol<<16;
		
		// ¦ܡ
		for( int i=0; i<(int)(RB60/sizeof(DWORD)+DWidth); i++ ) *(doff++) = BCol;
		
		// ݥ碌
		doff += (bbuf->pitch-LB60-P6WIDTH-RB60)/sizeof(DWORD) + DWidth;
		
		if( --HLine ) VOff = VOffL;				// ʣ饤ξVRAMեå᤹
		else          AOff = (AOff+0x20)&0x1ff;	// ȥӥ塼Ȥ512bytesǥ롼
	}
	
	// ¦ܡ
	for( int i=0; i<(int)(bbuf->pitch/sizeof(DWORD)*BB60); i++ ) *(doff++) = BCol;
}


// ----------------------------------
//     PC-6001mk2
// ----------------------------------

// ----------------------------------
//     [N60] Mode 1,2 sub
// ----------------------------------
void VDG62::Mode_12( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Mode_12]\n" );
	
	WORD AVOFF = 0;
	
	// ܡ顼ɤĤ֤
	SDL_FillRect( bbuf, NULL, COL_MODE1[4] );
	
	for( int y=0; y<16; y++ ){
		for( int x=0; x<32; x++ ){
			switch( GetAttr(AVOFF)&0x60 ){
			case 0x00:			// *** ⡼0 ***(̵)
			case 0x20:			// *** ⡼1 ***
				Putc_M1( x, y, AVOFF );
				break;
			case 0x40:			// *** ⡼2-2 ***(̵)
			case 0x60:			// *** ⡼2-1 ***
				Putc_M2( x, y, AVOFF );
				break;
			}
			AVOFF++;
		}
	}
}


// ----------------------------------
//     [N60] Mode 3,4 sub
// ----------------------------------
void VDG62::Mode_34( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Mode_34]\n" );
	
	int AOff = 0;	// ȥӥ塼ȥեå;
	int VOff = 0;	// VRAMեå
	
	DWORD *doff = (DWORD *)bbuf->pixels + TB60*bbuf->pitch/sizeof(DWORD) + LB60/sizeof(DWORD);
	// ХåХåեꥢ
	SDL_FillRect( bbuf, NULL, COL_MODE1[4] );
	
	for( int y=0; y<P6HEIGHT; y++ ){
		for( int x=0; x<(int)(P6WIDTH/sizeof(DWORD)/2); x++ ){
			DWORD d;
			BYTE b1,b2,b3,b4;
			BYTE Adat = GetAttr(AOff);
			int Css   = (Adat>>1)&1;		// 顼å
			int Gmode = (Adat>>2)&7;		// եå⡼(07)
			BYTE Vram = GetVram( VOff++ );	// VRAMǡ
			
			// ()ե˥塼åߤƤν
			if( !(Adat&0x80) ) Gmode = -1;
			
			switch( Gmode ){
			case -1:	//  ()ե˥塼å
				b1 = COL_MODE1[Css<<1];
				d  = BTODW( b1, b1, b1, b1 );
				*(doff++) = d;
				*(doff++) = d;
				break;
				
			case 0:		//  64 64 顼(̵)
			case 1:		// 128 96 顼(̵)
			case 2:		// 128 64 顼(̵)
			case 3:		// 128192 顼
				b1 = COL_MODE3[Css][(Vram>>6)&3];
				b2 = COL_MODE3[Css][(Vram>>4)&3];
				*(doff++) = BTODW( b1, b1, b2, b2 );
				b3 = COL_MODE3[Css][(Vram>>2)&3];
				b4 = COL_MODE3[Css][ Vram    &3];
				*(doff++) = BTODW( b3, b3, b4, b4 );
				break;
				
			case 4:		// 128 64 Υ(̵)
			case 5:		// 128192 Υ(̵)
			case 6:		// 128 96 Υ(̵)
			case 7:		// 256192 Υ
				if( Mode4Col ){		// 顼ɽ
					// 128192 顼
					Css += Mode4Col*2;
					b1 = COL_MODE3[Css][(Vram>>6)&3];
					b2 = COL_MODE3[Css][(Vram>>4)&3];
					*(doff++) = BTODW( b1, b1, b2, b2 );
					b3 = COL_MODE3[Css][(Vram>>2)&3];
					b4 = COL_MODE3[Css][ Vram    &3];
					*(doff++) = BTODW( b3, b3, b4, b4 );
				}else{
					// 256192 Υ
					b1 = COL_MODE4[Css][(Vram>>7)&1];
					b2 = COL_MODE4[Css][(Vram>>6)&1];
					b3 = COL_MODE4[Css][(Vram>>5)&1];
					b4 = COL_MODE4[Css][(Vram>>4)&1];
					*(doff++) = BTODW( b1, b2, b3, b4 );
					b1 = COL_MODE4[Css][(Vram>>3)&1];
					b2 = COL_MODE4[Css][(Vram>>2)&1];
					b3 = COL_MODE4[Css][(Vram>>1)&1];
					b4 = COL_MODE4[Css][ Vram    &1];
					*(doff++) = BTODW( b1, b2, b3, b4 );
				}
				break;
			}
			++AOff &= 0x1ff;	// ȥӥ塼Ȥ512bytesǥ롼
		}
		doff += (bbuf->pitch-P6WIDTH)/sizeof(DWORD);
	}
}


// [N60] ⡼2 1ʸPUT
void VDG62::Putc_M2( int x, int y, WORD addr )
{
	PRINTD3( GRP_LOG, "[VDG][62][Putc2_M2] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr,fdat;
	DWORD fg,bg;
	DWORD *doff;
	WORD soff;
	
	data = GetVram( addr );
	attr = GetAttr( addr );
	
	doff = (DWORD *)bbuf->pixels + LB60/sizeof(DWORD) + TB60*bbuf->pitch/sizeof(DWORD) + x*2 + y*bbuf->pitch*12/sizeof(DWORD);
	soff = 0x1000 + (data&0x3f)*16;	// ߥեȤ 1000H-13FFH
	fg   = COL_MODE2[(attr<<1)&4 | (data>>6)&3];
	bg   = COL_MODE2[8];
	for( int yc=0; yc<12; yc++ ){
		fdat    = GetFont1( soff++ );
		*doff++ = BTODW( fdat&0x80 ? fg : bg, fdat&0x40 ? fg : bg, fdat&0x20 ? fg : bg, fdat&0x10 ? fg : bg );
		*doff   = BTODW( fdat&0x08 ? fg : bg, fdat&0x04 ? fg : bg, fdat&0x02 ? fg : bg, fdat&0x01 ? fg : bg );
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// [mk2] ⡼1 1ʸPUT
void VDG62::Putc_M21( int x, int y, WORD addr )
{
	PRINTD3( GRP_LOG, "[VDG][62][Putc_M21] -> x:%d y:%d addr:%04X\n", x, y, addr );
	
	BYTE data,attr,fdat;
	DWORD fg,bg;
	DWORD *doff;
	WORD soff;
	
	data = GetVram( addr );
	attr = GetAttr( addr );
	doff = (DWORD *)bbuf->pixels + TB62*bbuf->pitch/sizeof(DWORD) + LB62/sizeof(DWORD) + x*2 + y*bbuf->pitch*10/sizeof(DWORD);
	soff = (data+((attr&0x80)?256:0))*16;
	fg   = COL_MOD21[attr&0x0f];
	bg   = COL_MOD21[((attr>>4)&0x07)|((Css2&1)<<3)];
	for( int yc=0; yc<10; yc++ ){
		fdat    = GetFont2( soff++ );
		*doff++ = BTODW( fdat&0x80 ? fg : bg, fdat&0x40 ? fg : bg, fdat&0x20 ? fg : bg, fdat&0x10 ? fg : bg );
		*doff   = BTODW( fdat&0x08 ? fg : bg, fdat&0x04 ? fg : bg, fdat&0x02 ? fg : bg, fdat&0x01 ? fg : bg );
		doff   += bbuf->pitch/sizeof(DWORD)-1;
	}
}


// ----------------------------------
//     [mk2] Mode 1,2 sub
// ----------------------------------
void VDG62::Mode2_1( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Mode2_1]\n" );
	
	WORD AVOFF = 0;
	
	// ܡ顼ɤĤ֤
	SDL_FillRect( bbuf, NULL, COL_MOD21[0] );
	
	for( int y=0; y<20; y++ ){
		for( int x=0; x<40; x++ ){
			Putc_M21( x, y, AVOFF++ );
		}
	}
}


// ----------------------------------
//     [mk2] Mode 3 sub
// ----------------------------------
void VDG62::Mode2_3( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Mode2_3]\n" );
	
	int voff=0;
	BYTE data,attr;
	BYTE b1,b2,b3,b4;
	DWORD *doff;
	
	SDL_FillRect( bbuf, NULL, COL_MOD23[0][0] );
	
	doff = (DWORD *)bbuf->pixels + TB62*bbuf->pitch/sizeof(DWORD) + LB62/sizeof(DWORD);
	
	for( int y=0; y<P2HEIGHT; y++ ){
		for( int x=0; x<(int)(P2WIDTH/sizeof(DWORD)/2); x++ ){
			data = GetAttr( voff );
			attr = GetVram( voff++ );
			b1 = COL_MOD23[Css3][((attr>>4)&0x0c)|((data>>6)&3)];
			b2 = COL_MOD23[Css3][((attr>>2)&0x0c)|((data>>4)&3)];
			*(doff++) = BTODW( b1, b1, b2, b2 );
			b3 = COL_MOD23[Css3][( attr    &0x0c)|((data>>2)&3)];
			b4 = COL_MOD23[Css3][((attr<<2)&0x0c)|( data    &3)];
			*(doff++) = BTODW( b3, b3, b4, b4 );
		}
		doff += (LB62+RB62)/sizeof(DWORD);
	}
}

// ----------------------------------
//     [mk2] Mode 4 sub
// ----------------------------------
void VDG62::Mode2_4( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Mode2_4]\n" );
	
	int voff=0;
	BYTE data,attr,c;
	BYTE b1,b2,b3,b4;
	DWORD *doff;
	
	SDL_FillRect( bbuf, NULL, COL_MOD24[0][0] );
	
	doff = (DWORD *)bbuf->pixels + TB62*bbuf->pitch/sizeof(DWORD) + LB62/sizeof(DWORD);
	c = (Css2<<3)|(Css1<<2);
	
	for( int y=0; y<P2HEIGHT; y++ ){
		for( int x=0; x<(int)(P2WIDTH/sizeof(DWORD)/2); x++ ){
			data = GetVram( voff );
			attr = GetAttr( voff++ );
			b1 = COL_MOD24[Css3][c|((data>>6)&2)|((attr>>7)&1)];
			b2 = COL_MOD24[Css3][c|((data>>5)&2)|((attr>>6)&1)];
			b3 = COL_MOD24[Css3][c|((data>>4)&2)|((attr>>5)&1)];
			b4 = COL_MOD24[Css3][c|((data>>3)&2)|((attr>>4)&1)];
			*(doff++) = BTODW( b1, b2, b3, b4 );
			b1 = COL_MOD24[Css3][c|((data>>2)&2)|((attr>>3)&1)];
			b2 = COL_MOD24[Css3][c|((data>>1)&2)|((attr>>2)&1)];
			b3 = COL_MOD24[Css3][c|( data    &2)|((attr>>1)&1)];
			b4 = COL_MOD24[Css3][c|((data<<1)&2)|( attr    &1)];
			*(doff++) = BTODW( b1, b2, b3, b4 );
		}
		doff += (LB62+RB62)/sizeof(DWORD);
	}
}


////////////////////////////////////////////////////////////////
// VRAM Ƥ ХåХåե
////////////////////////////////////////////////////////////////
void VDG60::Vram2Screen( void )
{
	PRINTD( GRP_LOG, "[VDG][60][Vram2Screen]s\n" );
	
	if( SDL_MUSTLOCK( bbuf ) ) SDL_LockSurface( bbuf );
	
	if( CrtDisp ){
		// ⡼Ƚ
		// ե˥塼åȥեåκߤϹθʤ
		// (ϡŪˤϲǽߤݤ)
		if( GetAttr(0)&0x80 ) Mode_34();	// եå
		else                  Mode_12();	// ե˥塼å
	}else{
		CrtKill();
	}
	
	if( SDL_MUSTLOCK( bbuf ) ) SDL_UnlockSurface( bbuf );
}

void VDG62::Vram2Screen( void )
{
	PRINTD( GRP_LOG, "[VDG][62][Vram2Screen]\n" );
	
	if( SDL_MUSTLOCK( bbuf ) ) SDL_LockSurface( bbuf );
	
	if( CrtDisp ){
		if( N60Win ){	// N60
			// ⡼Ƚ
			// ե˥塼åȥեåκߤϹθʤ
			// (ϡŪˤϲǽߤݤ)
			if( GetAttr(0)&0x80 ) Mode_34();	// եå
			else                  Mode_12();	// ե˥塼å
		}else{			// N60m
			if( Mk2CharMode )		// 饯⡼ɡ
				Mode2_1();			// ⡼1,2
			else{
				// եå
				if( Mk2GraphMode ) Mode2_3();	// ⡼3
				else               Mode2_4();	// ⡼4
			}
		}
	}else{
		CrtKill();
	}
	
	if( SDL_MUSTLOCK( bbuf ) ) SDL_UnlockSurface( bbuf );
}


////////////////////////////////////////////////////////////////
// VRAM ɥ쥹ڤؤ
// ( bit 1,2 ͭ)
//   ! ɥ쥹MainRam֥åƬΥեåȤȤ
////////////////////////////////////////////////////////////////
void VDG60::SetVramAddr( BYTE data )
{
	PRINTD1( GRP_LOG, "[VDG][SetVramAddr] -> %d : ", data );
	
	//	[00] A:C000H V:C200H
	//	[01] A:E000H V:E200H
	//	[10] A:8000H V:8200H
	//	[11] A:A000H V:A200H
	// ¥ɥ쥹 +0x8000
	AttrAddr = (((data+4)&6)<<12);
	VramAddr = AttrAddr + 0x0200;
	
	PRINTD2( GRP_LOG, "V:%04X A:%04X\n", VramAddr, AttrAddr );
}

void VDG62::SetVramAddr( BYTE data )
{
	PRINTD1( GRP_LOG, "[VDG][62][SetVramAddr] -> %d : ", data );
	
	if( !N60Win ){			// mk2 or 66
		//	[00] A:8000H V:8400H or A000H
		//	[01] A:C000H V:C400H or E000H
		//	[10] A:0000H V:0400H or 2000H
		//	[11] A:4000H V:4400H or 6000H
		AttrAddr = ((data+4)&6)<<13;
		if( Mk2CharMode )	// 饯⡼
			VramAddr = AttrAddr + 0x0400;
		else				// եå⡼
			VramAddr = AttrAddr + 0x2000;
	}else{					// 60
		//	[00] A:C000H V:C200H
		//	[01] A:E000H V:E200H
		//	[10] A:8000H V:8200H
		//	[11] A:A000H V:A200H
		// mk2ʹߤξϼ¥ɥ쥹˰
		AttrAddr = 0x8000 | (((data+4)&6)<<12);
		VramAddr = AttrAddr + 0x0200;
	}
	
	PRINTD2( GRP_LOG, "V:%04X A:%04X\n", VramAddr, AttrAddr );
}


////////////////////////////////////////////////////////////////
// ȹ礻
////////////////////////////////////////////////////////////////
void VDG62::SetCss( BYTE data )
{
	PRINTD1( GRP_LOG, "[VDG][62][SetCss] -> %02X : ", data );
	
	Css1 =  data    &1;
	Css2 = (data>>1)&1;
	Css3 = (data>>2)&1;
	
	PRINTD3( GRP_LOG, "1:%d 2:%d 3:%d\n", Css1, Css2, Css3 );
}


////////////////////////////////////////////////////////////////
// CRTȥ⡼
////////////////////////////////////////////////////////////////
void VDG62::SetCrtControler( BYTE data )
{
	PRINTD1( GRP_LOG, "[VDG][62][SetCrtControler] -> %02X\n", data );
	
	CrtCtrlPara = data;
	
	// VRAM ɥ쥹ѥ᡼ջ
	BYTE vrampara = N60Win ? ((AttrAddr&0x7fff)>>12)-4 : (AttrAddr>>13)-4;
	
	N60Win       = data&2 ? TRUE : FALSE;	// ɥ       TRUE:N60        FALSE:N60m
	Mk2CharMode  = data&4 ? TRUE : FALSE;	// mk2 ɽ⡼         TRUE:饯 FALSE:եå
	Mk2GraphMode = data&8 ? TRUE : FALSE;	// mk2 եå TRUE:160*200    FALSE:320*200
	
	PRINTD1( GRP_LOG, " DATA        :%02X\n", data );
	PRINTD1( GRP_LOG, " N60Win      :%s\n", N60Win       ? "N60"  : "N60m" );
	PRINTD1( GRP_LOG, " Mk2CharMode :%s\n", Mk2CharMode  ? "Char" : "Graph" );
	PRINTD1( GRP_LOG, " Mk2GraphMode:%d\n", Mk2GraphMode ? 160    : 320 );
	
	// VRAM ɥ쥹
	SetVramAddr( vrampara );
	
	// BUSREQѲ
	ChangeBUSREQ();
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
////////////////////////////////////////////////////////////////
BOOL VDG60::DokoSave( cIni *Ini )
{
	cSche::evinfo e;
	char stren[16];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "VDG", NULL, "VramAddr",	"0x%04X",		VramAddr );
	Ini->PutEntry( "VDG", NULL, "AttrAddr",	"0x%04X",		AttrAddr );
	
	Ini->PutEntry( "VDG", NULL, "CrtDisp",	"%s",			CrtDisp ? "Yes" : "No" );
	Ini->PutEntry( "VDG", NULL, "N60Win",	"%s",			N60Win ? "Yes" : "No" );
	
	Ini->PutEntry( "VDG", NULL, "VLines",	"%d",			VLines );
	Ini->PutEntry( "VDG", NULL, "VLcnt",	"%d",			VLcnt );
	Ini->PutEntry( "VDG", NULL, "HWait",	"%d",			HWait );
	
	
	// ٥
	e.id = EID_VSYNC;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "VDG", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	e.id = EID_HSYNC;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "VDG", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	return TRUE;
}

BOOL VDG62::DokoSave( cIni *Ini )
{
	cSche::evinfo e;
	char stren[16];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "VDG", NULL, "VramAddr",	"0x%04X",		VramAddr );
	Ini->PutEntry( "VDG", NULL, "AttrAddr",	"0x%04X",		AttrAddr );
	
	Ini->PutEntry( "VDG", NULL, "CrtDisp",	"%s",			CrtDisp ? "Yes" : "No" );
	Ini->PutEntry( "VDG", NULL, "N60Win",	"%s",			N60Win ? "Yes" : "No" );
	
	Ini->PutEntry( "VDG", NULL, "VLines",	"%d",			VLines );
	Ini->PutEntry( "VDG", NULL, "VLcnt",	"%d",			VLcnt );
	Ini->PutEntry( "VDG", NULL, "HWait",	"%d",			HWait );
	
	Ini->PutEntry( "VDG", NULL, "Mk2CharMode",	"%s",		Mk2CharMode ? "Yes" : "No" );
	Ini->PutEntry( "VDG", NULL, "Mk2GraphMode",	"%s",		Mk2GraphMode ? "Yes" : "No" );
	Ini->PutEntry( "VDG", NULL, "Css1",			"%d",		Css1 );
	Ini->PutEntry( "VDG", NULL, "Css2",			"%d",		Css2 );
	Ini->PutEntry( "VDG", NULL, "Css3",			"%d",		Css3 );
	Ini->PutEntry( "VDG", NULL, "CrtCtrlPara",	"0x%02X",	CrtCtrlPara );
	
	// ٥
	e.id = EID_VSYNC;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "VDG", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	e.id = EID_HSYNC;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "VDG", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɤǤLOAD
////////////////////////////////////////////////////////////////
BOOL VDG60::DokoLoad( cIni *Ini )
{
	int st,yn;
	cSche::evinfo e;
	char stren[16];
	char strrs[64];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->GetInt(   "VDG", "VramAddr",		&st,		VramAddr );	VramAddr = st;
	Ini->GetInt(   "VDG", "AttrAddr",		&st,		AttrAddr );	AttrAddr = st;
	
	Ini->GetTruth( "VDG", "CrtDisp",		&CrtDisp,	CrtDisp );
	Ini->GetTruth( "VDG", "N60Win",			&N60Win,	N60Win );
	
	Ini->GetInt(   "VDG", "VLines",			&st,		VLines );	VLines = st;
	Ini->GetInt(   "VDG", "VLcnt",			&st,		VLcnt );	VLcnt = st;
	Ini->GetInt(   "VDG", "HWait",			&st,		HWait );	HWait = st;
	
	// ٥
	e.id = EID_VSYNC;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "VDG", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	e.id = EID_HSYNC;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "VDG", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	return TRUE;
}

BOOL VDG62::DokoLoad( cIni *Ini )
{
	int st,yn;
	cSche::evinfo e;
	char stren[16];
	char strrs[64];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->GetInt(   "VDG", "VramAddr",		&st,		VramAddr );	VramAddr = st;
	Ini->GetInt(   "VDG", "AttrAddr",		&st,		AttrAddr );	AttrAddr = st;
	
	Ini->GetTruth( "VDG", "CrtDisp",		&CrtDisp,	CrtDisp );
	Ini->GetTruth( "VDG", "N60Win",			&N60Win,	N60Win );
	
	Ini->GetInt(   "VDG", "VLines",			&st,		VLines );	VLines = st;
	Ini->GetInt(   "VDG", "VLcnt",			&st,		VLcnt );	VLcnt = st;
	Ini->GetInt(   "VDG", "HWait",			&st,		HWait );	HWait = st;
	
	Ini->GetTruth( "VDG", "Mk2CharMode",	&Mk2CharMode,	Mk2CharMode );
	Ini->GetTruth( "VDG", "Mk2GraphMode",	&Mk2GraphMode,	Mk2GraphMode );
	Ini->GetInt(   "VDG", "Css1",			&Css1,			Css1 );
	Ini->GetInt(   "VDG", "Css2",			&Css2,			Css2 );
	Ini->GetInt(   "VDG", "Css3",			&Css3,			Css3 );
	Ini->GetInt(   "VDG", "CrtCtrlPara",	&st,			CrtCtrlPara );	CrtCtrlPara = st;
	
	// ٥
	e.id = EID_VSYNC;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "VDG", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	e.id = EID_HSYNC;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "VDG", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	return TRUE;
}



////////////////////////////////////////////////////////////////
//  device description
////////////////////////////////////////////////////////////////
const Device::Descriptor VDG60::descriptor = {
	VDG60::indef, VDG60::outdef
};

const Device::OutFuncPtr VDG60::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &VDG60::OutB0H )
};

const Device::InFuncPtr VDG60::indef[] = {
};


const Device::Descriptor VDG62::descriptor = {
	VDG62::indef, VDG62::outdef
};

const Device::OutFuncPtr VDG62::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &VDG62::OutB0H ),
	STATIC_CAST( Device::OutFuncPtr, &VDG62::OutC0H ),
	STATIC_CAST( Device::OutFuncPtr, &VDG62::OutC1H )
};

const Device::InFuncPtr VDG62::indef[] = {
	STATIC_CAST( Device::InFuncPtr, &VDG62::InC1H )
};
